# Copyright (c) 2025 GeoUtils developers
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module implementing operations on raster chunks, defines subfunctions called by "dask" and "multiproc" modules."""

from __future__ import annotations

from typing import Any, Literal, TypeVar

import geopandas as gpd
import numpy as np
import pandas as pd
import rasterio as rio
from rasterio import CRS

from geoutils._typing import NDArrayNum
from geoutils.projtools import _get_bounds_projected, _get_footprint_projected
from geoutils.raster._geotransformations import _rio_reproject

# REPROJECT (subfunctions called both in "dask" or "multiproc" module)

# At the date of April 2024: not supported by Rioxarray
# Part of the code (defining a GeoGrid and GeoTiling classes) is inspired by
# https://github.com/opendatacube/odc-geo/pull/88, modified to be concise, stand-alone and rely only on
# Rasterio/GeoPandas

# We define a GeoGrid and GeoTiling class (which composes GeoGrid) to consistently deal with georeferenced footprints
# of chunked grids
GeoGridType = TypeVar("GeoGridType", bound="GeoGrid")


class GeoGrid:
    """
    Georeferenced grid class.

    Describes a georeferenced grid through a geotransform (one-sided bounds and resolution), shape and CRS.
    """

    def __init__(self, transform: rio.transform.Affine, shape: tuple[int, int], crs: rio.crs.CRS | None):

        self._transform = transform
        self._shape = shape
        self._crs = crs

    @property
    def transform(self) -> rio.transform.Affine:
        return self._transform

    @property
    def crs(self) -> rio.crs.CRS:
        return self._crs

    @property
    def shape(self) -> tuple[int, int]:
        return self._shape

    @property
    def height(self) -> int:
        return self.shape[0]

    @property
    def width(self) -> int:
        return self.shape[1]

    @property
    def res(self) -> tuple[int, int]:
        return self.transform[0], abs(self.transform[4])

    def bounds_projected(self, crs: rio.crs.CRS = None) -> rio.coords.BoundingBox:
        if crs is None:
            crs = self.crs
        bounds = rio.coords.BoundingBox(*rio.transform.array_bounds(self.height, self.width, self.transform))
        return _get_bounds_projected(bounds=bounds, in_crs=self.crs, out_crs=crs)

    @property
    def bounds(self) -> rio.coords.BoundingBox:
        return self.bounds_projected()

    def footprint_projected(self, crs: rio.crs.CRS = None) -> gpd.GeoDataFrame:
        if crs is None:
            crs = self.crs
        return _get_footprint_projected(self.bounds, in_crs=self.crs, out_crs=crs, densify_points=100)

    @property
    def footprint(self) -> gpd.GeoDataFrame:
        return self.footprint_projected()

    @classmethod
    def from_dict(cls: type[GeoGridType], dict_meta: dict[str, Any]) -> GeoGridType:
        """Create a GeoGrid from a dictionary containing transform, shape and CRS."""
        return cls(**dict_meta)

    def translate(
        self: GeoGridType,
        xoff: float,
        yoff: float,
        distance_unit: Literal["georeferenced"] | Literal["pixel"] = "pixel",
    ) -> GeoGridType:
        """Translate into a new geogrid (not inplace)."""

        if distance_unit not in ["georeferenced", "pixel"]:
            raise ValueError("Argument 'distance_unit' should be either 'pixel' or 'georeferenced'.")

        # Get transform
        dx, b, xmin, d, dy, ymax = list(self.transform)[:6]

        # Convert pixel offsets to georeferenced units
        if distance_unit == "pixel":
            # Can either multiply the offset by the resolution
            # xoff *= self.res[0]
            # yoff *= self.res[1]

            # Or use the boundaries instead! (maybe less floating point issues? doesn't seem to matter in tests)
            xoff = xoff / self.shape[1] * (self.bounds.right - self.bounds.left)
            yoff = yoff / self.shape[0] * (self.bounds.top - self.bounds.bottom)

        shifted_transform = rio.transform.Affine(dx, b, xmin + xoff, d, dy, ymax + yoff)

        return self.from_dict({"transform": shifted_transform, "crs": self.crs, "shape": self.shape})


def _get_block_ids_per_chunk(chunks: tuple[tuple[int, ...], tuple[int, ...]]) -> list[dict[str, int]]:
    """Get location of chunks based on array shape and list of chunk sizes."""

    # Get number of chunks
    num_chunks = (len(chunks[0]), len(chunks[1]))

    # Get robust list of chunk locations (using what is done in block_id of dask.array.map_blocks)
    # https://github.com/dask/dask/blob/24493f58660cb933855ba7629848881a6e2458c1/dask/array/core.py#L908
    from dask.utils import cached_cumsum

    starts = [cached_cumsum(c, initial_zero=True) for c in chunks]
    nb_blocks = num_chunks[0] * num_chunks[1]
    ixi, iyi = np.unravel_index(np.arange(nb_blocks), shape=(num_chunks[0], num_chunks[1]))
    # Starting and ending indexes "s" and "e" for both X/Y, to place the chunk in the full array
    block_ids = [
        {
            "num_block": i,
            "ys": starts[0][ixi[i]],
            "xs": starts[1][iyi[i]],
            "ye": starts[0][ixi[i] + 1],
            "xe": starts[1][iyi[i] + 1],
        }
        for i in range(nb_blocks)
    ]

    return block_ids


class ChunkedGeoGrid:
    """
    Chunked georeferenced grid class.

    Associates a georeferenced grid to chunks (possibly of varying sizes).
    """

    def __init__(self, grid: GeoGrid, chunks: tuple[tuple[int, ...], tuple[int, ...]]):

        self._grid = grid
        self._chunks = chunks

    @property
    def grid(self) -> GeoGrid:
        return self._grid

    @property
    def chunks(self) -> tuple[tuple[int, ...], tuple[int, ...]]:
        return self._chunks

    def get_block_locations(self) -> list[dict[str, int]]:
        """Get block locations in 2D: xstart, xend, ystart, yend."""
        return _get_block_ids_per_chunk(self._chunks)

    def get_blocks_as_geogrids(self) -> list[GeoGrid]:
        """Get blocks as geogrids with updated transform/shape."""

        block_ids = self.get_block_locations()

        list_geogrids = []
        for bid in block_ids:
            # We get the block size
            block_shape = (bid["ye"] - bid["ys"], bid["xe"] - bid["xs"])
            # Build a temporary geogrid with the same transform as the full grid, but with the chunk shape
            geogrid_tmp = GeoGrid(transform=self.grid.transform, crs=self.grid.crs, shape=block_shape)
            # And shift it to the right location (X is positive in index direction, Y is negative)
            geogrid_block = geogrid_tmp.translate(xoff=bid["xs"], yoff=-bid["ys"])
            list_geogrids.append(geogrid_block)

        return list_geogrids

    def get_block_footprints(self, crs: rio.crs.CRS = None) -> gpd.GeoDataFrame:
        """Get block projected footprints as a single geodataframe."""

        geogrids = self.get_blocks_as_geogrids()
        footprints = [gg.footprint_projected(crs=crs) if crs is not None else gg.footprint for gg in geogrids]

        return pd.concat(footprints)


def _chunks2d_from_chunksizes_shape(
    chunksizes: tuple[int, int], shape: tuple[int, int]
) -> tuple[tuple[int, ...], tuple[int, ...]]:
    """Get tuples of chunk sizes for X/Y dimensions based on chunksizes and array shape."""

    # Chunksize is fixed, except for the last chunk depending on the shape
    chunks_y = tuple(
        min(
            chunksizes[0],
            shape[0] - i * chunksizes[0],
        )
        for i in range(int(np.ceil(shape[0] / chunksizes[0])))
    )
    chunks_x = tuple(
        min(
            chunksizes[1],
            shape[1] - i * chunksizes[1],
        )
        for i in range(int(np.ceil(shape[1] / chunksizes[1])))
    )

    return chunks_y, chunks_x


def _combined_blocks_shape_transform(
    sub_block_ids: list[dict[str, int]], src_geogrid: GeoGrid
) -> tuple[dict[str, Any], list[dict[str, int]]]:
    """Derive combined shape and transform from a subset of several blocks (for source input during reprojection)."""

    # Get combined shape by taking min of X/Y starting indices, max of X/Y ending indices
    all_xs, all_ys, all_xe, all_ye = ([b[s] for b in sub_block_ids] for s in ["xs", "ys", "xe", "ye"])
    minmaxs = {"min_xs": np.min(all_xs), "max_xe": np.max(all_xe), "min_ys": np.min(all_ys), "max_ye": np.max(all_ye)}
    combined_shape = (minmaxs["max_ye"] - minmaxs["min_ys"], minmaxs["max_xe"] - minmaxs["min_xs"])

    # Shift source transform with start indexes to get the one for combined block location
    combined_transform = src_geogrid.translate(xoff=minmaxs["min_xs"], yoff=-minmaxs["min_ys"]).transform

    # Compute relative block indexes that will be needed to reconstruct a square array in the delayed function,
    # by subtracting the minimum starting indices in X/Y
    relative_block_indexes = [
        {"r" + s1 + s2: b[s1 + s2] - minmaxs["min_" + s1 + "s"] for s1 in ["x", "y"] for s2 in ["s", "e"]}
        for b in sub_block_ids
    ]

    combined_meta = {"src_shape": combined_shape, "src_transform": tuple(combined_transform)}

    return combined_meta, relative_block_indexes


def _build_geotiling_and_meta(
    src_shape: tuple[int, int],
    src_transform: rio.transform.Affine,
    src_crs: CRS,
    dst_shape: tuple[int, int],
    dst_transform: rio.transform.Affine,
    dst_crs: CRS,
    src_chunks: tuple[tuple[int, ...], tuple[int, ...]],
    dst_chunksizes: tuple[int, int],
) -> tuple[
    ChunkedGeoGrid,
    ChunkedGeoGrid,
    tuple[tuple[int, ...], tuple[int, ...]],
    list[list[int]],
    list[dict[str, int]],
    list[tuple[dict[str, Any], list[dict[str, int]]]],
    list[GeoGrid],
]:
    """
    Constructs georeferenced tiling information and reprojection metadata for both source and destination grids,
    used to support block-wise reprojection operations (e.g. with multiprocessing or dask).

    This function performs the following:
    1. Constructs `GeoGrid` and `ChunkedGeoGrid` objects for source and destination rasters,
       based on provided shape, transform, CRS, and chunk sizes.
    2. Computes spatial footprints for each chunk in both grids, and determines which
       source chunks intersect each destination chunk (with a buffer to ensure overlap).
    3. For each destination chunk, calculates metadata required for reprojection, including:
       - The combined shape and transform of all intersecting source chunks.
       - The specific shape and transform of the destination block.

    :return: A tuple containing:
        - Source `ChunkedGeoGrid`
        - Destination `ChunkedGeoGrid`
        - Destination chunks
        - Mapping from destination to intersecting source block indices
        - Array of source block locations
        - List of metadata dictionaries per destination block
        - List of destination `GeoGrid` blocks
    """

    # 1/ Define source and destination chunked georeferenced grid through simple classes storing CRS/transform/shape,
    # which allow to consistently derive shape/transform for each block and their CRS-projected footprints

    # Define GeoGrids for source/destination array
    src_geogrid = GeoGrid(transform=src_transform, shape=src_shape, crs=src_crs)
    dst_geogrid = GeoGrid(transform=dst_transform, shape=dst_shape, crs=dst_crs)

    # Create tilings
    src_geotiling = ChunkedGeoGrid(grid=src_geogrid, chunks=src_chunks)
    dst_chunks = _chunks2d_from_chunksizes_shape(chunksizes=dst_chunksizes, shape=dst_shape)
    dst_geotiling = ChunkedGeoGrid(grid=dst_geogrid, chunks=dst_chunks)

    # 2/ Get footprints of tiles in CRS of destination array, with a buffer of 2 pixels for destination ones to ensure
    # overlap, then map indexes of source blocks that intersect a given destination block
    src_footprints = src_geotiling.get_block_footprints(crs=dst_crs)
    dst_footprints = dst_geotiling.get_block_footprints().buffer(2 * max(dst_geogrid.res))
    dest2source = [list(np.where(dst.intersects(src_footprints).values)[0]) for dst in dst_footprints]

    # 3/ To reconstruct a square source array during chunked reprojection, we need to derive the combined shape and
    # transform of each tuples of source blocks
    src_block_ids = src_geotiling.get_block_locations()
    meta_params = [
        (
            _combined_blocks_shape_transform(sub_block_ids=[src_block_ids[i] for i in sbid], src_geogrid=src_geogrid)
            if len(sbid) > 0
            else ({}, [])
        )
        for sbid in dest2source
    ]

    # Append dst shape/transform to metadata
    dst_block_geogrids = dst_geotiling.get_blocks_as_geogrids()
    for i, (c, _) in enumerate(meta_params):
        c.update({"dst_shape": dst_block_geogrids[i].shape, "dst_transform": tuple(dst_block_geogrids[i].transform)})

    return src_geotiling, dst_geotiling, dst_chunks, dest2source, src_block_ids, meta_params, dst_block_geogrids


def _reproject_per_block(
    *src_arrs: tuple[NDArrayNum], block_ids: list[dict[str, int]], combined_meta: dict[str, Any], **kwargs: Any
) -> NDArrayNum:
    """
    Reprojection per destination block (also rebuilds a square array combined from intersecting source blocks).
    """

    # If no source chunk intersects, we return a chunk of destination nodata values
    if len(src_arrs) == 0:
        # We can use float32 to return NaN, will be cast to other floating type later if that's not source array dtype
        dst_arr = np.zeros(combined_meta["dst_shape"], dtype=np.dtype("float32"))
        dst_arr[:] = kwargs["dst_nodata"]
        return dst_arr

    # First, we build an empty array with the combined shape, only with nodata values
    is_multiband = len(src_arrs[0].shape) > 2
    shape = (src_arrs[0].shape[0], *combined_meta["src_shape"]) if is_multiband else combined_meta["src_shape"]

    comb_src_arr = np.full(shape, kwargs["src_nodata"], dtype=src_arrs[0].dtype)

    # Then fill it with the source chunks values
    for arr, bid in zip(src_arrs, block_ids):
        comb_src_arr[..., bid["rys"] : bid["rye"], bid["rxs"] : bid["rxe"]] = arr

    # Now, we can simply call Rasterio!

    # We build the combined transform from tuple
    src_transform = rio.transform.Affine(*combined_meta["src_transform"])
    dst_transform = rio.transform.Affine(*combined_meta["dst_transform"])

    # Reproject wrapper

    # Force the number of threads to 1 to avoid Dask/Rasterio conflicting on multi-threading
    kwargs.update(
        {
            "dst_shape": combined_meta["dst_shape"],
            "src_transform": src_transform,
            "dst_transform": dst_transform,
            "num_threads": 1,
        }
    )
    # Define dtype if undefined
    if "dtype" not in kwargs:
        kwargs.update({"dtype": comb_src_arr.dtype})

    src_mask = comb_src_arr == kwargs["src_nodata"]
    dst_arr, dst_mask = _rio_reproject(src_arr=comb_src_arr, src_mask=src_mask, reproj_kwargs=kwargs)  # type: ignore
    if np.issubdtype(dst_arr.dtype, np.floating):
        dst_arr[dst_mask] = np.nan

    return dst_arr



"""Test tiling tools for arrays and rasters."""

import numpy as np
import pytest

import geoutils as gu
from geoutils import examples
from geoutils.raster.tiling import _generate_tiling_grid


class TestTiling:

    landsat_b4_path = examples.get_path_test("everest_landsat_b4")

    def test_subdivide_array(self) -> None:
        # Import optional scikit-image or skip test

        test_shape = (6, 4)
        test_count = 4
        subdivision_grid = gu.raster.subdivide_array(test_shape, test_count)

        assert subdivision_grid.shape == test_shape
        assert np.unique(subdivision_grid).size == test_count

        assert np.unique(gu.raster.subdivide_array((3, 3), 3)).size == 3

        with pytest.raises(ValueError, match=r"Expected a 2D shape, got 1D shape.*"):
            gu.raster.subdivide_array((5,), 2)

        with pytest.raises(ValueError, match=r"Shape.*smaller than.*"):
            gu.raster.subdivide_array((5, 2), 15)

    @pytest.mark.parametrize("overlap", [0, 5])  # type: ignore
    def test_tiling(self, overlap: int) -> None:

        # Test with mock data
        tiling_grid_mock = _generate_tiling_grid(0, 0, 100, 100, 50, 50, overlap)
        if overlap == 0:
            expected_tiling = np.array([[[0, 50, 0, 50], [0, 50, 50, 100]], [[50, 100, 0, 50], [50, 100, 50, 100]]])
            assert np.array_equal(tiling_grid_mock, expected_tiling)
        elif overlap == 5:
            expected_tiling = np.array(
                [
                    [[0, 55, 0, 55], [0, 55, 45, 100]],
                    [[45, 100, 0, 55], [45, 100, 45, 100]],
                ]
            )
            assert np.array_equal(tiling_grid_mock, expected_tiling)

        tiling_grid_mock = _generate_tiling_grid(0, 0, 55, 55, 50, 50, overlap)
        if overlap == 0:
            expected_tiling = np.array([[[0, 50, 0, 50], [0, 50, 50, 55]], [[50, 55, 0, 50], [50, 55, 50, 55]]])
            assert np.array_equal(tiling_grid_mock, expected_tiling)
        elif overlap == 5:
            expected_tiling = np.array([[[0, 55, 0, 55]]])
            assert np.array_equal(tiling_grid_mock, expected_tiling)

        # Test with real data
        img = gu.Raster(self.landsat_b4_path)

        # Define tiling parameters
        row_split, col_split = 100, 100
        row_max, col_max = img.shape

        # Generate the tiling grid
        tiling_grid = _generate_tiling_grid(0, 0, row_max, col_max, row_split, col_split, overlap)

        # Calculate expected number of tiles
        nb_row_tiles = np.ceil(row_max / row_split).astype(int)
        nb_col_tiles = np.ceil(col_max / col_split).astype(int)

        if 0 < col_max % col_split <= overlap:
            nb_col_tiles = max(nb_col_tiles - 1, 1)
        if 0 < row_max % row_split <= overlap:
            nb_row_tiles = max(nb_row_tiles - 1, 1)

        # Check that the tiling grid has the expected shape
        assert tiling_grid.shape == (nb_row_tiles, nb_col_tiles, 4)

        # Check the boundaries of the first and last tile
        assert np.array_equal(
            tiling_grid[0, 0],
            np.array(
                [
                    0,
                    min(row_split + overlap, row_max),
                    0,
                    min(col_split + overlap, col_max),
                ]
            ),
        )
        assert np.array_equal(
            tiling_grid[-1, -1],
            np.array(
                [
                    (nb_row_tiles - 1) * row_split - overlap,
                    row_max,
                    (nb_col_tiles - 1) * col_split - overlap,
                    col_max,
                ]
            ),
        )

        # Check if overlap is consistent between tiles
        for row in range(nb_row_tiles - 1):
            assert tiling_grid[row + 1, 0, 0] == tiling_grid[row, 0, 1] - 2 * overlap

        for col in range(nb_col_tiles - 1):
            assert tiling_grid[0, col + 1, 2] == tiling_grid[0, col, 3] - 2 * overlap

    def test_tiling_overlap_errors(self) -> None:
        with pytest.raises(ValueError):
            _generate_tiling_grid(0, 0, 100, 100, 50, 50, -1)
        with pytest.raises(TypeError):
            _generate_tiling_grid(0, 0, 100, 100, 50, 50, 0.5)  # type: ignore



"""Test for PointCloud class."""

from __future__ import annotations

import os
import re
import tempfile
import warnings
from importlib.util import find_spec

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pytest
from geopandas.testing import assert_geodataframe_equal
from pyproj import CRS
from shapely import Polygon

import geoutils as gu
from geoutils import PointCloud
from geoutils._typing import NDArrayNum

DO_PLOT = False


class TestPointCloud:

    # 1/ Synthetic 2D points with main column and no auxiliary column
    rng = np.random.default_rng(42)
    arr_points = rng.integers(low=1, high=1000, size=(100, 3)) + rng.normal(0, 0.15, size=(100, 3))
    gdf1 = gpd.GeoDataFrame(
        data={"b1": arr_points[:, 2]}, geometry=gpd.points_from_xy(x=arr_points[:, 0], y=arr_points[:, 1]), crs=4326
    )

    # 2/ Synthetic 2D points with main column and with auxiliary column
    arr_points2 = rng.integers(low=1, high=1000, size=(100, 4)) + rng.normal(0, 0.15, size=(100, 4))
    gdf2 = gpd.GeoDataFrame(
        data=arr_points2[:, 2:],
        columns=["b1", "b2"],
        geometry=gpd.points_from_xy(x=arr_points2[:, 0], y=arr_points2[:, 1]),
        crs=4326,
    )

    # 3/ Synthetic 3D points and with auxiliary column
    rng = np.random.default_rng(42)
    arr_points = rng.integers(low=1, high=1000, size=(100, 3)) + rng.normal(0, 0.15, size=(100, 3))
    gdf3 = gpd.GeoDataFrame(
        data=[],
        columns=[],
        geometry=gpd.points_from_xy(x=arr_points[:, 0], y=arr_points[:, 1], z=arr_points[:, 2]),
        crs=4326,
    )
    # To test 3D with an extra column
    gdf32 = gpd.GeoDataFrame(
        data=arr_points2[:, 3:],
        columns=["b2"],
        geometry=gpd.points_from_xy(x=arr_points[:, 0], y=arr_points[:, 1], z=arr_points[:, 2]),
        crs=4326,
    )

    # 4/ LAS file
    fn_las = gu.examples.get_path_test("coromandel_lidar")

    # 5/ Non-point vector (for error raising)
    poly = Polygon([(5, 5), (6, 5), (6, 6), (5, 6)])
    gdf4 = gpd.GeoDataFrame({"geometry": [poly]}, crs="EPSG:4326")

    def test_init(self) -> None:
        """Test instantiation of a point cloud."""

        # 1/ For a single column point cloud with 2D geometries
        pc = PointCloud(self.gdf1, data_column="b1")

        # Assert that both the dataframe and data column name are equal
        assert pc.data_column == "b1"
        assert_geodataframe_equal(pc.ds, self.gdf1)

        # 2/ For a point cloud with 3D geometries, no need to pass a data column
        pc = PointCloud(self.gdf3)

        # Assert that both the dataframe and data column name are equal
        assert pc.data_column is None
        assert_geodataframe_equal(pc.ds, self.gdf3)

    def test_init_las(self) -> None:
        # Import optional laspy or skip test
        pytest.importorskip("laspy")

        # 1/ For a point cloud from LAS/LAZ file, no need to pass a data column to get the default Z
        pc = PointCloud(self.fn_las)

        assert pc.data_column == "Z"
        assert not pc.is_loaded

        # 2/ But we can still specify another one
        pc = PointCloud(self.fn_las, data_column="number_of_returns")

        assert pc.data_column == "number_of_returns"
        assert not pc.is_loaded

    @pytest.mark.skipif(find_spec("laspy") is not None, reason="Only runs if laspy is missing.")  # type: ignore
    def test_init_las__missing_dep(self) -> None:
        """Check that an absent laspy dependency raises the proper error."""

        with pytest.raises(ImportError, match="Optional dependency 'laspy' required.*"):
            PointCloud(self.fn_las)

    def test_init__errors(self) -> None:
        """Test errors raised by point cloud instantiation."""

        # If the data column does not exist
        with pytest.raises(ValueError, match="Data column column_that_does_not_exist not found*"):
            PointCloud(self.gdf1, data_column="column_that_does_not_exist")

        # If vector is not only comprised of points
        with pytest.raises(ValueError, match="This vector file contains non-point geometries*"):
            PointCloud(self.gdf4, data_column="z")

        # If the data column is not defined and the geometries are 2D
        with pytest.raises(ValueError, match="A data column name must be passed for a point cloud with 2D.*"):
            PointCloud(self.gdf1)

    def test_load__las(self) -> None:
        """
        Test loading of a point cloud (only possible with a LAS file).

        This test also serves to test the overridden methods "crs", "bounds", "nb_points", "nongeo_columns" in relation
        to loading.
        """

        # Import optional laspy or skip test
        pytest.importorskip("laspy")

        # 1/ Check unloaded and loaded attributes are all the same

        pc = PointCloud(self.fn_las)

        # Check point cloud is not loaded, and fetch before metadata
        assert not pc.is_loaded
        before_crs = pc.crs
        before_bounds = pc.bounds
        before_nb_points = pc.point_count
        before_columns = pc.columns

        # Load and fetch after metadata
        pc.load(columns="all")
        assert pc.is_loaded

        after_crs = pc.crs
        after_bounds = pc.bounds
        after_nb_points = pc.point_count
        after_columns = pc.columns

        # Check those are equal
        assert before_crs == after_crs
        assert before_bounds == after_bounds
        assert before_nb_points == after_nb_points
        assert before_columns == after_columns

        # 2/ Check default column argument
        pc = PointCloud(self.fn_las)
        pc.load()

        assert pc._nongeo_columns == ["Z"]

        # 3/ Check implicit loading when calling a function requiring .ds
        pc = PointCloud(self.fn_las)
        assert not pc.is_loaded

        pc.buffer(distance=0.1)
        assert pc.is_loaded

    def test_load__errors(self) -> None:
        """Test errors raised by loading."""

        # Import optional laspy or skip test
        pytest.importorskip("laspy")

        pc = PointCloud(self.fn_las)
        pc.load()

        # Error if already loaded
        with pytest.raises(ValueError, match="Data are already loaded."):
            pc.load()

        pc = PointCloud(self.fn_las)
        pc._name = None
        with pytest.raises(AttributeError, match="Cannot load as filename is not set anymore.*"):
            pc.load()

    def test_getitem_setitem(self) -> None:
        """Test the __getitem__ method ([]) for indexing and __setitem__ for index assignment."""

        # -- First, we test mask or boolean array indexing and assignment, specific to point clouds --

        # Open a point cloud
        # We need to do a deep copy to avoid modifying the original object
        pc = PointCloud(self.gdf1.copy(deep=True), data_column="b1")
        pc_orig = pc.copy()

        # Create a boolean array of the same shape, and a mask of the same transform/crs
        rng = np.random.default_rng(42)
        arr = rng.integers(low=0, high=2, size=pc.point_count, dtype=bool)
        mask = gu.PointCloud.from_xyz(x=pc.geometry.x.values, y=pc.geometry.y.values, z=arr, crs=pc.crs)

        # Check that indexing works with both of those
        vals_arr = pc[arr]
        vals_mask = pc[mask]

        # Those indexing operations should yield the same point cloud
        assert isinstance(vals_arr, PointCloud)
        assert isinstance(vals_mask, PointCloud)
        assert vals_mask.pointcloud_equal(vals_arr)

        # Check for 3D point cloud

        pc3d = PointCloud(self.gdf3.copy(deep=True))
        vals_arr = pc3d[arr]
        vals_mask = pc3d[mask]

        assert isinstance(vals_arr, PointCloud)
        assert isinstance(vals_mask, PointCloud)
        assert vals_mask.pointcloud_equal(vals_arr)

        # 2/ Now, we test index assignment
        pc2 = pc_orig.copy()

        # It should work with a number, or a 1D array of the same length as the indexed one
        pc["b1"] = 1.0
        pc2["b1"] = np.ones(pc2.point_count)

        # The point clouds should be the same
        assert pc2.pointcloud_equal(pc)

        # It should also work with a boolean point cloud of the same length as the indexes one
        pc3 = pc_orig.copy()
        pc3[pc3 > 500] = np.nan
        assert all(np.isnan(pc3.data[pc3.data > 500]))

        # It should work for 3D point inputs
        pc3d = PointCloud(self.gdf3.copy(deep=True))
        pc3d[pc3d > 500] = np.nan
        assert all(np.isnan(pc3d.data[pc3d.data > 500]))

        # -- Finally, we check that errors are raised for both indexing and index assignment --

        # For indexing
        op_name_index = "an indexing operation"
        message_pc = "Both point clouds must have the same points X/Y coordinates and CRS for {}."

        # An error when the georeferencing of the Mask does not match
        mask.translate(1, 1, inplace=True)
        with pytest.raises(ValueError, match=re.escape(message_pc.format(op_name_index))):
            pc[mask]

    def test_data_column(self) -> None:
        """Test the setting and getting of the main data column."""

        # Assert column is set properly at instantiation
        pc = PointCloud(self.gdf1, data_column="b1")
        assert pc.data_column == "b1"
        assert np.array_equal(pc.data, self.gdf1["b1"].values)

        # And can be reset to another name if it exists
        pc2 = PointCloud(self.gdf2, data_column="b1")
        assert pc2.data_column == "b1"
        assert np.array_equal(pc2.data, self.gdf2["b1"].values)

        # First syntax
        pc2.data_column = "b2"
        assert pc2.data_column == "b2"
        assert np.array_equal(pc2.data, self.gdf2["b2"].values)

        # Equivalent syntax
        pc2.set_data_column("b1")
        assert pc2.data_column == "b1"
        assert np.array_equal(pc2.data, self.gdf2["b1"].values)

        # Assert no data column is set for 3D points, using the Z coordinates instead
        pc3 = PointCloud(self.gdf3)
        assert pc3.data_column is None
        assert np.array_equal(pc3.data, self.gdf3.geometry.z.values)

    def test_data_column__errors(self) -> None:
        """Test errors raised during setting of data column."""

        pc = PointCloud(self.gdf1, data_column="b1")
        # If the data column does not exist
        with pytest.raises(ValueError, match="Data column column_that_does_not_exist not found*"):
            pc.data_column = "column_that_does_not_exist"
        # Equivalent syntax
        with pytest.raises(ValueError, match="Data column column_that_does_not_exist not found*"):
            pc.set_data_column("column_that_does_not_exist")

        # If a data column name is passed for 3D points
        with pytest.warns(UserWarning, match="Overriding 3D points with*"):
            pc4 = PointCloud(self.gdf32, data_column="b2")
            assert pc4.data_column == "b2"
            assert np.array_equal(pc4.data, self.gdf32["b2"].values)

    def test_data(self) -> None:
        """Test the setting and getting of the main data, depending on input geometry."""

        # For a point cloud using 2D geometries + a main data column
        pc = PointCloud(self.gdf1, data_column="b1")
        assert np.array_equal(pc.data, self.gdf1["b1"].values)
        pc += 1
        assert np.array_equal(pc.data, self.gdf1["b1"].values + 1)

        # For a point cloud using 3D geometries
        pc2 = PointCloud(self.gdf3)
        assert np.array_equal(pc2.data, self.gdf3.geometry.z.values)
        pc2 += 1
        assert np.array_equal(pc2.data, self.gdf3.geometry.z.values + 1)

    def test_from_array(self) -> None:
        """Test building point cloud from array."""

        # Build from array and compare
        pc1 = PointCloud(self.gdf1, data_column="b1")
        pc_from_arr = PointCloud.from_array(self.arr_points, crs=4326, data_column="b1")
        assert pc_from_arr.pointcloud_equal(pc1)

        # Should be the same with transposed array
        pc_from_arr = PointCloud.from_array(self.arr_points.T, crs=4326, data_column="b1")
        assert pc_from_arr.pointcloud_equal(pc1)

    def test_from_array__errors(self) -> None:
        """Test errors raised during creation with array."""

        array = np.ones((4, 5))
        with pytest.raises(ValueError, match="Array must be of shape 3xN or Nx3."):
            PointCloud.from_array(array, crs=4326)

    def test_from_tuples(self) -> None:
        """Test building point cloud from list of tuples."""

        pc1 = PointCloud(self.gdf1, data_column="b1")
        tuples_xyz = list(zip(self.arr_points[:, 0], self.arr_points[:, 1], self.arr_points[:, 2]))
        pc_from_tuples = PointCloud.from_tuples(tuples_xyz=tuples_xyz, crs=4326, data_column="b1")
        assert pc_from_tuples.pointcloud_equal(pc1)

    def test_from_xyz(self) -> None:
        """Test building point cloud from xyz array-like."""

        # Build from array and compare
        pc1 = PointCloud(self.gdf1, data_column="b1")
        pc_from_xyz = PointCloud.from_xyz(
            x=self.arr_points[:, 0], y=self.arr_points[:, 1], z=self.arr_points[:, 2], crs=4326, data_column="b1"
        )
        assert pc_from_xyz.pointcloud_equal(pc1)

        # Test with lists
        pc_from_xyz = PointCloud.from_xyz(
            x=list(self.arr_points[:, 0]),
            y=list(self.arr_points[:, 1]),
            z=list(self.arr_points[:, 2]),
            crs=4326,
            data_column="b1",
        )
        assert pc_from_xyz.pointcloud_equal(pc1)

        # Test with tuples
        pc_from_xyz = PointCloud.from_xyz(
            x=tuple(self.arr_points[:, 0]),
            y=tuple(self.arr_points[:, 1]),
            z=tuple(self.arr_points[:, 2]),
            crs=4326,
            data_column="b1",
        )
        assert pc_from_xyz.pointcloud_equal(pc1)

        # Build with the use_z option to compare to 3D points
        pc3 = PointCloud(self.gdf3)
        pc3_from_xyz = PointCloud.from_xyz(
            x=self.arr_points[:, 0], y=self.arr_points[:, 1], z=self.arr_points[:, 2], crs=4326, use_z=True
        )
        assert pc3_from_xyz.pointcloud_equal(pc3)

    def test_to_array(self) -> None:
        """Test exporting point cloud to array."""

        # Convert point cloud and compare
        pc1 = PointCloud(self.gdf1, data_column="b1")
        arr_from_pc = pc1.to_array()
        assert np.array_equal(arr_from_pc, self.arr_points.T)

    def test_to_tuples(self) -> None:
        """Test exporting point cloud to tuples."""

        # Convert point cloud and compare
        pc1 = PointCloud(self.gdf1, data_column="b1")
        tuples_xyz = list(zip(self.arr_points[:, 0], self.arr_points[:, 1], self.arr_points[:, 2]))
        tuples_from_pc = pc1.to_tuples()
        assert tuples_from_pc == tuples_xyz

    def test_to_xyz(self) -> None:
        """Test exporting point cloud to xyz arrays."""

        # Convert point cloud and compare
        pc1 = PointCloud(self.gdf1, data_column="b1")
        xyz_from_pc = pc1.to_xyz()
        assert np.array_equal(np.stack(xyz_from_pc), self.arr_points.T)

    specific_method_args = {
        "reproject": {"crs": CRS.from_epsg(32610)},
        "crop": {
            "crop_geom": (
                np.min(gdf1.geometry.x.values),
                np.min(gdf1.geometry.y.values),
                np.max(gdf1.geometry.x.values),
                np.max(gdf1.geometry.y.values),
            )
        },
        "translate": {"xoff": 1, "yoff": 1},
        "set_precision": {"grid_size": 1},
        "to_crs": {"crs": CRS.from_epsg(32610)},
        "set_crs": {"crs": CRS.from_epsg(32610), "allow_override": True},
        "rename_geometry": {"col": "lol"},
    }

    def test_to_las(self) -> None:

        # Import optional laspy or skip test
        pytest.importorskip("laspy")

        # 1/ For a X/Y/Z point cloud with no auxiliary data
        pc1 = PointCloud(self.gdf1, data_column="b1")

        # Temporary folder
        temp_dir = tempfile.TemporaryDirectory()

        # Save file to temporary file, with defaults opts
        temp_file = os.path.join(temp_dir.name, "test.las")
        pc1.to_las(temp_file)
        saved1 = gu.PointCloud(temp_file)

        # Check loaded point cloud is equal to saved one
        atol = 0.1
        assert np.allclose(pc1.geometry.x.values, saved1.geometry.x.values, atol=atol)
        assert np.allclose(pc1.geometry.y.values, saved1.geometry.y.values, atol=atol)
        assert np.allclose(pc1.data, saved1.data, atol=atol)

        # 2/ For a X/Y/Z point cloud with some auxiliary data
        pc2 = PointCloud(self.gdf2, data_column="b1")

        # Save file to temporary file, with defaults opts
        temp_file = os.path.join(temp_dir.name, "test2.las")
        pc2.to_las(temp_file)
        saved2 = gu.PointCloud(temp_file)
        saved2.load(["Z", "b2"])

        # Check loaded point cloud is equal to saved one
        assert np.allclose(pc2.geometry.x.values, saved2.geometry.x.values, atol=atol)
        assert np.allclose(pc2.geometry.y.values, saved2.geometry.y.values, atol=atol)
        assert np.allclose(pc2.data, saved2.data, atol=atol)
        assert np.allclose(pc2["b2"].values, saved2["b2"].values, atol=atol)

    @pytest.mark.skipif(find_spec("laspy") is not None, reason="Only runs if laspy is missing.")  # type: ignore
    def test_to_las__missing_dep(self) -> None:
        """Check to_las() raises the proper error if laspy is not installed."""

        with pytest.raises(ImportError, match="Optional dependency 'laspy' required.*"):
            PointCloud(self.fn_las)

    @pytest.mark.parametrize(
        "method", ["reproject", "crop", "translate", "set_precision", "to_crs", "set_crs", "rename_geometry"]
    )  # type: ignore
    def test_cast_vector_methods__geometry_invariant(self, method: str) -> None:
        """Test that method that don't modify geometry do cast back to a PointCloud."""

        pc1 = PointCloud(self.gdf1, data_column="b1")

        getattr(pc1, method)(**self.specific_method_args[method])
        assert isinstance(pc1, PointCloud)


class TestArithmetic:
    """
    Test that all arithmetic overloading functions work as expected.
    """

    # Create fake point clouds with random values in 0-255 and dtype uint8
    # TODO: Add the case where a mask exists in the array, as in test_data_setter
    rng = np.random.default_rng(42)
    nb_points = 20

    min_val = np.iinfo("uint8").min
    max_val = np.iinfo("uint8").max
    coords = rng.integers(min_val, max_val, (2, nb_points), dtype="uint8")
    data1 = rng.integers(min_val, max_val, nb_points, dtype="uint8")
    data2 = rng.integers(min_val, max_val, nb_points, dtype="uint8")

    # Create random masks
    mask1 = rng.integers(0, 2, size=nb_points, dtype=bool)
    mask2 = rng.integers(0, 2, size=nb_points, dtype=bool)
    mask3 = rng.integers(0, 2, size=nb_points, dtype=bool)

    crs = CRS(4326)
    wrong_crs = CRS(32610)

    # Create point clouds
    pc1 = gu.PointCloud.from_xyz(x=coords[0, :], y=coords[1, :], z=data1, crs=crs)
    pc2 = gu.PointCloud.from_xyz(x=coords[0, :], y=coords[1, :], z=data2, crs=crs)

    # Tests with different dtype
    pc1_f32 = gu.PointCloud.from_xyz(x=coords[0, :], y=coords[1, :], z=data1.astype("float32"), crs=crs)

    # With zero for operations such as division
    data_zeros = np.copy(data2)
    data_zeros[0:2] = 0
    pc2_zero = gu.PointCloud.from_xyz(x=coords[0, :], y=coords[1, :], z=data_zeros, crs=crs)

    pc1_wrong_crs = gu.PointCloud.from_xyz(x=coords[0, :], y=coords[1, :], z=data1, crs=wrong_crs)

    # Create rasters with different shape, crs or transforms for testing errors
    # Wrong shaped arrays to check errors are raised
    arr_wrong_shape = rng.integers(min_val, max_val, (3, nb_points - 1), dtype="int32") + rng.normal(
        size=(3, nb_points - 1)
    )
    mask_wrong_shape = rng.integers(0, 2, size=nb_points - 1, dtype=bool)

    # Wrong coordinate array to check errors are raised
    arr_wrong_coord = rng.integers(min_val, max_val, (3, nb_points), dtype="int32") + rng.normal(size=(3, nb_points))
    pc1_wrong_shape = gu.PointCloud.from_array(arr_wrong_shape, crs=crs)

    pc1_wrong_coords = gu.PointCloud.from_array(arr_wrong_coord, crs=crs)

    def test_pointcloud_equal(self) -> None:
        """
        Test that pointcloud_equal() works as expected.
        """
        pc1 = self.pc1
        pc2 = pc1.copy()
        assert pc1.pointcloud_equal(pc2)

        # Change data
        pc2.data = pc2.data + 1
        assert not pc1.pointcloud_equal(pc2)

        # Change dtype
        pc2 = pc1.copy()
        pc2 = pc2.astype("float32")
        assert not pc1.pointcloud_equal(pc2)

        # Change coordinates
        pc2 = pc1.copy()
        pc2 = pc2.translate(2, 1)
        assert not pc1.pointcloud_equal(pc2)

        # Change CRS
        pc2 = pc1.copy()
        pc2 = pc2.set_crs(self.wrong_crs, allow_override=True)
        assert not pc1.pointcloud_equal(pc2)

    def test_georeferenced_coords_equal(self) -> None:
        """
        Test that equal for shape, crs and transform work as expected
        """

        # -- Test 1: based on a copy --
        pc1 = self.pc1
        pc2 = pc1.copy()
        assert pc1.georeferenced_coords_equal(pc2)

        # Change data
        pc2.data = pc2.data + 1
        assert pc1.georeferenced_coords_equal(pc2)

        # Change dtype
        pc2 = pc1.copy()
        pc2 = pc2.astype("float32")
        assert pc1.georeferenced_coords_equal(pc2)

        # Change coords
        pc2 = pc1.copy()
        pc2 = pc2.translate(2, 1)
        assert not pc1.georeferenced_coords_equal(pc2)

        # Change CRS
        pc2 = pc1.copy()
        pc2 = pc2.set_crs(self.wrong_crs, allow_override=True)
        assert not pc1.georeferenced_coords_equal(pc2)

        # -- Test 2: based on another Raster with one different georeferenced grid attribute --

        assert not pc1.georeferenced_coords_equal(self.pc1_wrong_crs)

        assert not pc1.georeferenced_coords_equal(self.pc1_wrong_shape)

        assert not pc1.georeferenced_coords_equal(self.pc1_wrong_coords)

    # List of operations with two operands
    ops_2args = [
        "__add__",
        "__radd__",
        "__sub__",
        "__rsub__",
        "__mul__",
        "__rmul__",
        "__truediv__",
        "__rtruediv__",
        "__floordiv__",
        "__rfloordiv__",
        "__mod__",
    ]

    @pytest.mark.parametrize("op", ops_2args)  # type: ignore
    def test_ops_2args_expl(self, op: str) -> None:
        """
        Check that arithmetic overloading functions, with two operands, work as expected when called explicitly.
        """
        warnings.filterwarnings("ignore", message="invalid value encountered")

        # Test various inputs: Point clouds with different dtypes, np.ndarray, single number
        pc1 = self.pc1
        pc1_f32 = self.pc1_f32
        pc2 = self.pc2
        pc2_zero = self.pc2_zero
        rng = np.random.default_rng(42)
        array = rng.integers(1, 255, self.nb_points).astype("float64")
        floatval = 3.14
        intval = 1

        # Test with 2 uint8 point clouds
        pc1 = self.pc1
        pc2 = self.pc2
        pc3 = getattr(pc1, op)(pc2)
        ctype = np.promote_types(pc1.data.dtype, pc2.data.dtype)
        numpy_output = getattr(pc1.data.astype(ctype), op)(pc2.data.astype(ctype))
        assert isinstance(pc3, gu.PointCloud)
        assert np.all(pc3.data == numpy_output)
        assert pc3.data.dtype == numpy_output.dtype
        assert pc3.crs == pc1.crs

        # Test original data are not modified
        pc1_copy = pc1.copy()
        pc2_copy = pc2.copy()
        pc3 = getattr(pc1, op)(pc2)
        assert isinstance(pc3, gu.PointCloud)
        assert pc1.pointcloud_equal(pc1_copy)
        assert pc2.pointcloud_equal(pc2_copy)

        # Test with different dtypes
        pc1 = self.pc1_f32
        pc2 = self.pc2
        pc3 = getattr(pc1_f32, op)(pc2)
        assert pc3.data.dtype == np.dtype("float32")
        assert np.all(pc3.data == getattr(pc1.data, op)(pc2.data))

        # Test with zeros values (e.g. division)
        pc1 = self.pc1
        pc3 = getattr(pc1, op)(pc2_zero)
        assert np.all(pc3.data == getattr(pc1.data, op)(pc2_zero.data))

        # Test with a numpy array
        pc1 = self.pc1_f32
        pc3 = getattr(pc1, op)(array)
        assert isinstance(pc3, gu.PointCloud)
        assert np.all(pc3.data == getattr(pc1.data, op)(array))

        # Test with an integer
        pc3 = getattr(pc1, op)(intval)
        assert isinstance(pc3, gu.PointCloud)
        assert np.all(pc3.data == getattr(pc1.data, op)(intval))

        # Test with a float value
        pc3 = getattr(pc1, op)(floatval)
        assert isinstance(pc3, gu.PointCloud)
        # Behaviour is more complex for scalars since NumPy 2.0,
        # so simply comparing it is consistent with that of masked arrays
        assert pc3.pointcloud_equal(self.copy(getattr(pc1.data, op)(floatval), pc_ref=pc3))

    reflective_ops = [["__add__", "__radd__"], ["__mul__", "__rmul__"]]

    @pytest.mark.parametrize("ops", reflective_ops)  # type: ignore
    def test_reflectivity(self, ops: list[str]) -> None:
        """
        Check reflective operations
        """
        warnings.filterwarnings("ignore", message="invalid value encountered")

        # Test various inputs: Raster with different dtypes, np.ndarray, single number
        rng = np.random.default_rng(42)
        array = rng.integers(1, 255, self.nb_points).astype("float64")
        floatval = 3.14
        intval = 1

        # Get reflective operations
        op1, op2 = ops

        # Test with uint8 rasters
        pc3 = getattr(self.pc1, op1)(self.pc2)
        pc4 = getattr(self.pc1, op2)(self.pc2)
        assert pc3.pointcloud_equal(pc4)

        # Test with different dtypes
        pc3 = getattr(self.pc1_f32, op1)(self.pc2)
        pc4 = getattr(self.pc1_f32, op2)(self.pc2)
        assert pc3.pointcloud_equal(pc4)

        # Test with zeros values (e.g. division)
        pc3 = getattr(self.pc1, op1)(self.pc2_zero)
        pc4 = getattr(self.pc1, op2)(self.pc2_zero)
        assert pc3.pointcloud_equal(pc4)

        # Test with a numpy array
        pc3 = getattr(self.pc1, op1)(array)
        pc4 = getattr(self.pc1, op2)(array)
        assert pc3.pointcloud_equal(pc4)

        # Test with an integer
        pc3 = getattr(self.pc1, op1)(intval)
        pc4 = getattr(self.pc1, op2)(intval)
        assert pc3.pointcloud_equal(pc4)

        # Test with a float value
        pc3 = getattr(self.pc1, op1)(floatval)
        pc4 = getattr(self.pc1, op2)(floatval)
        assert pc3.pointcloud_equal(pc4)

    @classmethod
    def copy(
        cls: type[TestArithmetic],
        data: NDArrayNum,
        pc_ref: gu.PointCloud,
    ) -> gu.PointCloud:
        """
        Generate a pointcloud from numpy array, with set georeferencing. Used for testing only.
        """

        return pc_ref.copy(new_array=data)

    def test_ops_2args_implicit(self) -> None:
        """
        Test certain arithmetic overloading when called with symbols (+, -, *, /, //, %).
        """
        warnings.filterwarnings("ignore", message="invalid value encountered")

        # Test various inputs: Raster with different dtypes, np.ndarray with 2D or 3D shape, single number
        pc1 = self.pc1
        pc1_f32 = self.pc1_f32
        pc2 = self.pc2
        rng = np.random.default_rng(42)
        array_2d = rng.integers(1, 255, (1, self.nb_points)).astype("uint8")
        array_1d = rng.integers(1, 255, self.nb_points).astype("uint8")
        floatval = 3.14

        # Addition
        assert (pc1 + pc2).pointcloud_equal(self.copy(pc1.data + pc2.data, pc_ref=pc1))
        assert (pc1_f32 + pc2).pointcloud_equal(self.copy(pc1_f32.data + pc2.data, pc_ref=pc1))
        assert (array_2d + pc2).pointcloud_equal(self.copy(array_2d + pc2.data, pc_ref=pc2))
        assert (pc2 + array_2d).pointcloud_equal(self.copy(pc2.data + array_2d, pc_ref=pc2))
        assert (array_1d + pc2).pointcloud_equal(self.copy(array_1d[np.newaxis, :] + pc2.data, pc_ref=pc2))
        assert (pc2 + array_1d).pointcloud_equal(self.copy(pc2.data + array_1d[np.newaxis, :], pc_ref=pc2))
        assert (pc1 + floatval).pointcloud_equal(self.copy(pc1.data + floatval, pc_ref=pc1))
        assert (floatval + pc1).pointcloud_equal(self.copy(floatval + pc1.data, pc_ref=pc1))
        assert (pc1 + pc2).pointcloud_equal(pc2 + pc1)

        # Multiplication
        assert (pc1 * pc2).pointcloud_equal(self.copy(pc1.data * pc2.data, pc_ref=pc1))
        assert (pc1_f32 * pc2).pointcloud_equal(self.copy(pc1_f32.data * pc2.data, pc_ref=pc1))
        assert (array_2d * pc2).pointcloud_equal(self.copy(array_2d * pc2.data, pc_ref=pc2))
        assert (pc2 * array_2d).pointcloud_equal(self.copy(pc2.data * array_2d, pc_ref=pc2))
        assert (array_1d * pc2).pointcloud_equal(self.copy(array_1d[np.newaxis, :] * pc2.data, pc_ref=pc2))
        assert (pc2 * array_1d).pointcloud_equal(self.copy(pc2.data * array_1d[np.newaxis, :], pc_ref=pc2))
        assert (pc1 * floatval).pointcloud_equal(self.copy(pc1.data * floatval, pc_ref=pc1))
        assert (floatval * pc1).pointcloud_equal(self.copy(floatval * pc1.data, pc_ref=pc1))
        assert (pc1 * pc2).pointcloud_equal(pc2 * pc1)

        # Subtraction
        assert (pc1 - pc2).pointcloud_equal(self.copy(pc1.data - pc2.data, pc_ref=pc1))
        assert (pc1_f32 - pc2).pointcloud_equal(self.copy(pc1_f32.data - pc2.data, pc_ref=pc1))
        assert (array_2d - pc2).pointcloud_equal(self.copy(array_2d - pc2.data, pc_ref=pc2))
        assert (pc2 - array_2d).pointcloud_equal(self.copy(pc2.data - array_2d, pc_ref=pc2))
        assert (array_1d - pc2).pointcloud_equal(self.copy(array_1d[np.newaxis, :] - pc2.data, pc_ref=pc2))
        assert (pc2 - array_1d).pointcloud_equal(self.copy(pc2.data - array_1d[np.newaxis, :], pc_ref=pc2))
        assert (pc1 - floatval).pointcloud_equal(self.copy(pc1.data - floatval, pc_ref=pc1))
        assert (floatval - pc1).pointcloud_equal(self.copy(floatval - pc1.data, pc_ref=pc1))

        # True division
        assert (pc1 / pc2).pointcloud_equal(self.copy(pc1.data / pc2.data, pc_ref=pc1))
        assert (pc1_f32 / pc2).pointcloud_equal(self.copy(pc1_f32.data / pc2.data, pc_ref=pc1))
        assert (array_2d / pc2).pointcloud_equal(self.copy(array_2d / pc2.data, pc_ref=pc2))
        assert (pc2 / array_2d).pointcloud_equal(self.copy(pc2.data / array_2d, pc_ref=pc2))
        assert (array_1d / pc2).pointcloud_equal(self.copy(array_1d[np.newaxis, :] / pc2.data, pc_ref=pc1))
        assert (pc2 / array_1d).pointcloud_equal(self.copy(pc2.data / array_1d[np.newaxis, :], pc_ref=pc2))
        assert (pc1 / floatval).pointcloud_equal(self.copy(pc1.data / floatval, pc_ref=pc1))
        assert (floatval / pc1).pointcloud_equal(self.copy(floatval / pc1.data, pc_ref=pc1))

        # Floor division
        assert (pc1 // pc2).pointcloud_equal(self.copy(pc1.data // pc2.data, pc_ref=pc1))
        assert (pc1_f32 // pc2).pointcloud_equal(self.copy(pc1_f32.data // pc2.data, pc_ref=pc1))
        assert (array_2d // pc2).pointcloud_equal(self.copy(array_2d // pc2.data, pc_ref=pc1))
        assert (pc2 // array_2d).pointcloud_equal(self.copy(pc2.data // array_2d, pc_ref=pc1))
        assert (array_1d // pc2).pointcloud_equal(self.copy(array_1d[np.newaxis, :] // pc2.data, pc_ref=pc1))
        assert (pc2 // array_1d).pointcloud_equal(self.copy(pc2.data // array_1d[np.newaxis, :], pc_ref=pc1))
        assert (pc1 // floatval).pointcloud_equal(self.copy(pc1.data // floatval, pc_ref=pc1))
        assert (floatval // pc1).pointcloud_equal(self.copy(floatval // pc1.data, pc_ref=pc1))

        # Modulo
        assert (pc1 % pc2).pointcloud_equal(self.copy(pc1.data % pc2.data, pc_ref=pc1))
        assert (pc1_f32 % pc2).pointcloud_equal(self.copy(pc1_f32.data % pc2.data, pc_ref=pc1))
        assert (array_2d % pc2).pointcloud_equal(self.copy(array_2d % pc2.data, pc_ref=pc1))
        assert (pc2 % array_2d).pointcloud_equal(self.copy(pc2.data % array_2d, pc_ref=pc1))
        assert (array_1d % pc2).pointcloud_equal(self.copy(array_1d[np.newaxis, :] % pc2.data, pc_ref=pc1))
        assert (pc2 % array_1d).pointcloud_equal(self.copy(pc2.data % array_1d[np.newaxis, :], pc_ref=pc1))
        assert (pc1 % floatval).pointcloud_equal(self.copy(pc1.data % floatval, pc_ref=pc1))

    def test_ops_logical_implicit(self) -> None:
        """
        Test logical arithmetic overloading when called with symbols (==, !=, <, <=, >, >=).
        """
        warnings.filterwarnings("ignore", message="invalid value encountered")

        # Test various inputs: Raster with different dtypes, np.ndarray with 2D or 3D shape, single number
        pc1 = self.pc1
        pc1_f32 = self.pc1_f32
        pc2 = self.pc2
        rng = np.random.default_rng(42)
        array_2d = rng.integers(1, 255, (1, self.nb_points)).astype("uint8")
        array_1d = rng.integers(1, 255, self.nb_points).astype("uint8")
        floatval = 3.14

        # Equality
        assert (pc1 == pc2).pointcloud_equal(self.copy(pc1.data == pc2.data, pc_ref=pc1))
        assert (pc1_f32 == pc2).pointcloud_equal(self.copy(pc1_f32.data == pc2.data, pc_ref=pc1))
        assert (array_2d == pc2).pointcloud_equal(self.copy(array_2d == pc2.data, pc_ref=pc2))
        assert (pc2 == array_2d).pointcloud_equal(self.copy(pc2.data == array_2d, pc_ref=pc2))
        assert (array_1d == pc2).pointcloud_equal(self.copy(array_1d[np.newaxis, :] == pc2.data, pc_ref=pc2))
        assert (pc2 == array_1d).pointcloud_equal(self.copy(pc2.data == array_1d[np.newaxis, :], pc_ref=pc2))
        assert (pc1 == floatval).pointcloud_equal(self.copy(pc1.data == floatval, pc_ref=pc1))
        assert (floatval == pc1).pointcloud_equal(self.copy(floatval == pc1.data, pc_ref=pc1))
        assert (pc1 == pc2).pointcloud_equal(pc2 == pc1)

        # Non-equality
        assert (pc1 != pc2).pointcloud_equal(self.copy(pc1.data != pc2.data, pc_ref=pc1))
        assert (pc1_f32 != pc2).pointcloud_equal(self.copy(pc1_f32.data != pc2.data, pc_ref=pc1))
        assert (array_2d != pc2).pointcloud_equal(self.copy(array_2d != pc2.data, pc_ref=pc2))
        assert (pc2 != array_2d).pointcloud_equal(self.copy(pc2.data != array_2d, pc_ref=pc2))
        assert (array_1d != pc2).pointcloud_equal(self.copy(array_1d[np.newaxis, :] != pc2.data, pc_ref=pc2))
        assert (pc2 != array_1d).pointcloud_equal(self.copy(pc2.data != array_1d[np.newaxis, :], pc_ref=pc2))
        assert (pc1 != floatval).pointcloud_equal(self.copy(pc1.data != floatval, pc_ref=pc1))
        assert (floatval != pc1).pointcloud_equal(self.copy(floatval != pc1.data, pc_ref=pc1))
        assert (pc1 != pc2).pointcloud_equal(pc2 != pc1)

        # Lower than
        assert (pc1 < pc2).pointcloud_equal(self.copy(pc1.data < pc2.data, pc_ref=pc1))
        assert (pc1_f32 < pc2).pointcloud_equal(self.copy(pc1_f32.data < pc2.data, pc_ref=pc1))
        assert (array_2d < pc2).pointcloud_equal(self.copy(array_2d < pc2.data, pc_ref=pc2))
        assert (pc2 < array_2d).pointcloud_equal(self.copy(pc2.data < array_2d, pc_ref=pc2))
        assert (array_1d < pc2).pointcloud_equal(self.copy(array_1d[np.newaxis, :] < pc2.data, pc_ref=pc2))
        assert (pc2 < array_1d).pointcloud_equal(self.copy(pc2.data < array_1d[np.newaxis, :], pc_ref=pc2))
        assert (pc1 < floatval).pointcloud_equal(self.copy(pc1.data < floatval, pc_ref=pc1))
        assert (floatval < pc1).pointcloud_equal(self.copy(floatval < pc1.data, pc_ref=pc1))

        # Lower equal
        assert (pc1 <= pc2).pointcloud_equal(self.copy(pc1.data <= pc2.data, pc_ref=pc1))
        assert (pc1_f32 <= pc2).pointcloud_equal(self.copy(pc1_f32.data <= pc2.data, pc_ref=pc1))
        assert (array_2d <= pc2).pointcloud_equal(self.copy(array_2d <= pc2.data, pc_ref=pc2))
        assert (pc2 <= array_2d).pointcloud_equal(self.copy(pc2.data <= array_2d, pc_ref=pc2))
        assert (array_1d <= pc2).pointcloud_equal(self.copy(array_1d[np.newaxis, :] <= pc2.data, pc_ref=pc1))
        assert (pc2 <= array_1d).pointcloud_equal(self.copy(pc2.data <= array_1d[np.newaxis, :], pc_ref=pc2))
        assert (pc1 <= floatval).pointcloud_equal(self.copy(pc1.data <= floatval, pc_ref=pc1))
        assert (floatval <= pc1).pointcloud_equal(self.copy(floatval <= pc1.data, pc_ref=pc1))

        # Greater than
        assert (pc1 > pc2).pointcloud_equal(self.copy(pc1.data > pc2.data, pc_ref=pc1))
        assert (pc1_f32 > pc2).pointcloud_equal(self.copy(pc1_f32.data > pc2.data, pc_ref=pc1))
        assert (array_2d > pc2).pointcloud_equal(self.copy(array_2d > pc2.data, pc_ref=pc1))
        assert (pc2 > array_2d).pointcloud_equal(self.copy(pc2.data > array_2d, pc_ref=pc1))
        assert (array_1d > pc2).pointcloud_equal(self.copy(array_1d[np.newaxis, :] > pc2.data, pc_ref=pc1))
        assert (pc2 > array_1d).pointcloud_equal(self.copy(pc2.data > array_1d[np.newaxis, :], pc_ref=pc1))
        assert (pc1 > floatval).pointcloud_equal(self.copy(pc1.data > floatval, pc_ref=pc1))
        assert (floatval > pc1).pointcloud_equal(self.copy(floatval > pc1.data, pc_ref=pc1))

        # Greater equal
        assert (pc1 >= pc2).pointcloud_equal(self.copy(pc1.data >= pc2.data, pc_ref=pc1))
        assert (pc1_f32 >= pc2).pointcloud_equal(self.copy(pc1_f32.data >= pc2.data, pc_ref=pc1))
        assert (array_2d >= pc2).pointcloud_equal(self.copy(array_2d >= pc2.data, pc_ref=pc1))
        assert (pc2 >= array_2d).pointcloud_equal(self.copy(pc2.data >= array_2d, pc_ref=pc1))
        assert (array_1d >= pc2).pointcloud_equal(self.copy(array_1d[np.newaxis, :] >= pc2.data, pc_ref=pc1))
        assert (pc2 >= array_1d).pointcloud_equal(self.copy(pc2.data >= array_1d[np.newaxis, :], pc_ref=pc1))
        assert (pc1 >= floatval).pointcloud_equal(self.copy(pc1.data >= floatval, pc_ref=pc1))

    def test_ops_logical_bitwise_implicit(self) -> None:
        # Create two masks
        pc1 = self.pc1
        m1 = self.pc1 > 128
        m2 = self.pc2 > 128
        rng = np.random.default_rng(42)
        array_1d = rng.integers(1, 255, self.nb_points).astype("uint8") > 128

        # Bitwise or
        assert (m1 | m2).pointcloud_equal(self.copy(m1.data | m2.data, pc_ref=pc1))
        assert (m1 | array_1d).pointcloud_equal(self.copy(m1.data | array_1d, pc_ref=pc1))
        assert (array_1d | m1).pointcloud_equal(self.copy(array_1d | m1.data, pc_ref=pc1))

        # Bitwise and
        assert (m1 & m2).pointcloud_equal(self.copy(m1.data & m2.data, pc_ref=pc1))
        assert (m1 & array_1d).pointcloud_equal(self.copy(m1.data & array_1d, pc_ref=pc1))
        assert (array_1d & m1).pointcloud_equal(self.copy(array_1d & m1.data, pc_ref=pc1))

        # Bitwise xor
        assert (m1 ^ m2).pointcloud_equal(self.copy(m1.data ^ m2.data, pc_ref=pc1))
        assert (m1 ^ array_1d).pointcloud_equal(self.copy(m1.data ^ array_1d, pc_ref=pc1))
        assert (array_1d ^ m1).pointcloud_equal(self.copy(array_1d ^ m1.data, pc_ref=pc1))

        # Bitwise invert
        assert (~m1).pointcloud_equal(self.copy(~m1.data, pc_ref=pc1))

    @pytest.mark.parametrize("op", ops_2args)  # type: ignore
    def test_raise_errors(self, op: str) -> None:
        """
        Test that errors are properly raised in certain situations.

        !! Important !! Here we test errors with the operator on the raster only (arithmetic overloading),
        calling with array first is supported with the NumPy interface and tested in ArrayInterface.
        """
        # Rasters with different CRS, transform, or shape
        # Different shape
        expected_message = (
            "Both point clouds must have the same points X/Y coordinates and CRS for an arithmetic operation."
        )
        with pytest.raises(ValueError, match=re.escape(expected_message)):
            getattr(self.pc2, op)(self.pc1_wrong_shape)

        # Different CRS
        with pytest.raises(ValueError, match=re.escape(expected_message)):
            getattr(self.pc2, op)(self.pc1_wrong_crs)

        # Different transform
        with pytest.raises(ValueError, match=re.escape(expected_message)):
            getattr(self.pc2, op)(self.pc1_wrong_coords)

        # Array with different shape
        expected_message = (
            "The array must be 1-dimensional with the same number of points as the point cloud for an arithmetic "
            "operation."
        )

        # Different shape, normal array with NaNs
        with pytest.raises(ValueError, match=re.escape(expected_message)):
            getattr(self.pc2, op)(self.pc1_wrong_shape.data)

        # Wrong type of "other"
        expected_message = "Operation between an object of type .* and a point cloud impossible."
        with pytest.raises(NotImplementedError, match=expected_message):
            getattr(self.pc1, op)("some_string")

    @pytest.mark.parametrize("power", [2, 3.14, -1])  # type: ignore
    def test_power(self, power: float | int) -> None:
        if power > 0:  # Integers to negative integer powers are not allowed.
            assert self.pc1**power == self.copy(self.pc1.data**power, pc_ref=self.pc1)
        assert self.pc1_f32**power == self.copy(self.pc1_f32.data**power, pc_ref=self.pc1_f32)

    @pytest.mark.parametrize("dtype", ["float32", "uint8", "int32"])  # type: ignore
    def test_numpy_functions(self, dtype: str) -> None:
        """Test how rasters can be used as/with numpy arrays."""
        warnings.simplefilter("error")

        # Create an array of unique values starting at 0 and ending at 24
        array = np.arange(25, dtype=dtype)
        # Create an associated dummy coordinates
        coords = np.zeros((2, 25))

        # Create a raster from the array
        pc = gu.PointCloud.from_xyz(x=coords[0, :], y=coords[1, :], z=array, crs=4326)

        # Test some ufuncs
        assert np.median(pc) == 12.0
        assert np.mean(pc) == 12.0

        # Check that rasters don't become arrays when using simple arithmetic.
        assert isinstance(pc + 1, gu.PointCloud)

        # Test the data setter method by creating a new array
        pc.data = array + 2

        # Check that the median updated accordingly.
        assert np.median(pc) == 14.0

        # Test
        pc += array

        assert isinstance(pc, gu.PointCloud)
        assert np.median(pc) == 26.0

    def test_plot(self) -> None:
        """Test the pointcloud plot."""

        # Create a dummy array of unique values and the associated coordinates
        array = np.arange(25, dtype=int)
        coords_x = [i for _ in range(5) for i in range(5)]
        coords_y = [i for i in range(5) for _ in range(5)]

        # Create the corresponding pointcloud
        pc = gu.PointCloud.from_xyz(x=coords_x, y=coords_y, z=array, crs=4326)

        # Test default plot
        pc.plot()
        if DO_PLOT:
            plt.show()
        else:
            plt.close()
        assert True

        # Test with new figure
        plt.figure()
        pc.plot()
        if DO_PLOT:
            plt.show()
        else:
            plt.close()
        assert True

        # Test with provided ax
        ax = plt.subplot(111)
        pc.plot(ax=ax)
        if DO_PLOT:
            plt.show()
        else:
            plt.close()
        assert True

        # Test vmin, vmax and cbar_title
        ax = plt.subplot(111)
        pc.plot(cmap="gray", vmin=0, vmax=20, cbar_title="Custom cbar", ax=ax)
        if DO_PLOT:
            plt.show()
        else:
            plt.close()
        assert True

        # Test save fig
        temp_dir = tempfile.TemporaryDirectory()
        temp_file = os.path.join(temp_dir.name, "test.png")
        pc.plot(savefig_fname=temp_file)
        if DO_PLOT:
            plt.show()
        else:
            plt.close()
        assert os.path.isfile(temp_file)


class TestArrayInterface:
    """Test that the array interface of PointCloud works as expected for ufuncs and array functions"""

    # -- First, we list all universal NumPy functions, or "ufuncs" --

    # All universal functions of NumPy, about 90 in 2022. See list: https://numpy.org/doc/stable/reference/ufuncs.html
    ufuncs_str = [
        ufunc
        for ufunc in np._core.umath.__all__
        if (
            ufunc[0] != "_"
            and ufunc.islower()
            and "err" not in ufunc
            and ufunc not in ["e", "pi", "frompyfunc", "euler_gamma", "vecdot", "vecmat", "matvec", "matmul"]
        )
    ]

    # Universal functions with one input argument and one output, corresponding to (in NumPy 1.22.4):
    # ['absolute', 'arccos', 'arccosh', 'arcsin', 'arcsinh', 'arctan', 'arctanh', 'cbrt', 'ceil', 'conj', 'conjugate',
    # 'cos', 'cosh', 'deg2rad', 'degrees', 'exp', 'exp2', 'expm1', 'fabs', 'floor', 'invert', 'isfinite', 'isinf',
    # 'isnan', 'isnat', 'log', 'log10', 'log1p', 'log2', 'logical_not', 'negative', 'positive', 'rad2deg', 'radians',
    # 'reciprocal', 'rint', 'sign', 'signbit', 'sin', 'sinh', 'spacing', 'sqrt', 'square', 'tan', 'tanh', 'trunc']
    ufuncs_str_1nin_1nout = [
        ufunc for ufunc in ufuncs_str if (getattr(np, ufunc).nin == 1 and getattr(np, ufunc).nout == 1)
    ]

    # Universal functions with one input argument and two output (Note: none exist for three outputs or above)
    # Those correspond to: ['frexp', 'modf']
    ufuncs_str_1nin_2nout = [
        ufunc for ufunc in ufuncs_str if (getattr(np, ufunc).nin == 1 and getattr(np, ufunc).nout == 2)
    ]

    # Universal functions with two input arguments and one output, corresponding to:
    # ['add', 'arctan2', 'bitwise_and', 'bitwise_or', 'bitwise_xor', 'copysign', 'divide', 'equal', 'floor_divide',
    #  'float_power', 'fmax', 'fmin', 'fmod', 'gcd', 'greater', 'greater_equal', 'heaviside', 'hypot', 'lcm', 'ldexp',
    #  'left_shift', 'less', 'less_equal', 'logaddexp', 'logaddexp2', 'logical_and', 'logical_or', 'logical_xor',
    #  'maximum', 'minimum', 'mod', 'multiply', 'nextafter', 'not_equal', 'power', 'remainder', 'right_shift',
    #  'subtract', 'true_divide']
    ufuncs_str_2nin_1nout = [
        ufunc for ufunc in ufuncs_str if (getattr(np, ufunc).nin == 2 and getattr(np, ufunc).nout == 1)
    ]

    # Universal functions with two input arguments and two outputs (Note: none exist for three outputs or above)
    # These correspond to: ['divmod']
    ufuncs_str_2nin_2nout = [
        ufunc for ufunc in ufuncs_str if (getattr(np, ufunc).nin == 2 and getattr(np, ufunc).nout == 2)
    ]

    # -- Second, we list array functions we intend to support in the array interface --

    # To my knowledge, there is no list that includes all numpy functions (and we probably don't want to test them all)
    # Let's include manually the important ones:
    # - statistics: normal and for NaNs;
    # - sorting and counting;
    # Most other math functions are already universal functions

    # Separate between two lists (single input and double input) for testing
    handled_functions_2in = gu.pointcloud.pointcloud._HANDLED_FUNCTIONS_2NIN
    handled_functions_1in = gu.pointcloud.pointcloud._HANDLED_FUNCTIONS_1NIN

    # Details below:
    # NaN functions: [f for f in np.lib.nanfunctions.__all__]
    # nanstatfuncs = ['nansum', 'nanmax', 'nanmin', 'nanargmax', 'nanargmin', 'nanmean', 'nanmedian', 'nanpercentile',
    #             'nanvar', 'nanstd', 'nanprod', 'nancumsum', 'nancumprod', 'nanquantile']

    # Statistics and sorting matching NaN functions: https://numpy.org/doc/stable/reference/routines.statistics.html
    # and https://numpy.org/doc/stable/reference/routines.sort.html
    # statfuncs = ['sum', 'max', 'min', 'argmax', 'argmin', 'mean', 'median', 'percentile', 'var', 'std', 'prod',
    #              'cumsum', 'cumprod', 'quantile']

    # Sorting and counting ounting with single array input:
    # sortfuncs = ['sort', 'count_nonzero', 'unique]

    # --  Third, we define the test data --

    # We create two random array of varying dtype
    nb_points = 20
    min_val = np.iinfo("int32").min
    max_val = np.iinfo("int32").max
    rng = np.random.default_rng(42)
    # All points cloud should have the same X/Y for operations to work, we only change the data
    coords = rng.integers(min_val, max_val, (2, nb_points), dtype="int32") + rng.normal(size=(2, nb_points))
    data1 = rng.integers(min_val, max_val, nb_points, dtype="int32") + rng.normal(size=nb_points)
    data2 = rng.integers(min_val, max_val, nb_points, dtype="int32") + rng.normal(size=nb_points)
    # This third one is to try ufunc methods like reduce()
    data3 = rng.integers(min_val, max_val, nb_points, dtype="int32") + rng.normal(size=nb_points)

    # Create random masks
    mask1 = rng.integers(0, 2, size=nb_points, dtype=bool)
    mask2 = rng.integers(0, 2, size=nb_points, dtype=bool)
    mask3 = rng.integers(0, 2, size=nb_points, dtype=bool)

    # Assert that there is at least one unmasked value
    assert np.count_nonzero(~mask1) > 0
    assert np.count_nonzero(~mask2) > 0
    assert np.count_nonzero(~mask3) > 0

    # CRS
    crs = CRS(4326)

    # Wrong shaped arrays to check errors are raised
    arr_wrong_shape = rng.integers(min_val, max_val, (3, nb_points - 1), dtype="int32") + rng.normal(
        size=(3, nb_points - 1)
    )
    mask_wrong_shape = rng.integers(0, 2, size=nb_points - 1, dtype=bool)

    # Wrong coordinate array to check errors are raised
    arr_wrong_coord = rng.integers(min_val, max_val, (3, nb_points), dtype="int32") + rng.normal(size=(3, nb_points))

    # Wrong CRS input
    wrong_crs = CRS(32610)

    @pytest.mark.parametrize("ufunc_str", ufuncs_str_1nin_1nout + ufuncs_str_1nin_2nout)  # type: ignore
    @pytest.mark.parametrize("dtype", ["uint8", "int8", "float32"])  # type: ignore
    def test_array_ufunc_1nin_1nout(self, ufunc_str: str, dtype: str) -> None:
        """Test that ufuncs with one input and one output consistently return the same result as for masked arrays."""

        data = self.data1.astype(dtype)
        # If floating type, create NaNs
        if np.issubdtype(dtype, np.floating):
            data[self.mask1] = np.nan
        pc = gu.PointCloud.from_xyz(x=self.coords[0], y=self.coords[1], z=data, crs=self.crs)

        # Get ufunc
        ufunc = getattr(np, ufunc_str)

        # Find the common dtype between the point cloud and the most constrained input type (first character is input)
        try:
            com_dtype = np.promote_types(dtype, ufunc.types[0][0])
        # The promote_types function raises an error for object dtypes (previously returned by find_common_dtypes)
        # (TypeError needed for backwards compatibility; also exceptions.DTypePromotionError for NumPy 1.25 and above)
        except TypeError:
            com_dtype = np.dtype("O")

        # Catch warnings
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)

            print(pc.data)

            # Check if our input dtype is possible on this ufunc, if yes check that outputs are identical
            if com_dtype in [str(np.dtype(t[0])) for t in ufunc.types]:  # noqa
                # For a single output
                if ufunc.nout == 1:
                    assert np.array_equal(ufunc(pc.data), ufunc(pc).data, equal_nan=True)

                # For two outputs
                elif ufunc.nout == 2:
                    outputs_pc = ufunc(pc)
                    output_arr = ufunc(pc.data)
                    assert np.array_equal(output_arr[0], outputs_pc[0].data, equal_nan=True) and np.array_equal(
                        output_arr[1], outputs_pc[1].data, equal_nan=True
                    )

            # If the input dtype is not possible, check that NumPy raises a TypeError
            else:
                with pytest.raises(TypeError):
                    ufunc(pc.data)
                with pytest.raises(TypeError):
                    ufunc(pc)

    @pytest.mark.parametrize("ufunc_str", ufuncs_str_2nin_1nout + ufuncs_str_2nin_2nout)  # type: ignore
    @pytest.mark.parametrize("dtype1", ["uint8", "int8", "float32"])  # type: ignore
    @pytest.mark.parametrize("dtype2", ["uint8", "int8", "float32"])  # type: ignore
    def test_array_ufunc_2nin_1nout(self, ufunc_str: str, dtype1: str, dtype2: str) -> None:
        """Test that ufuncs with two input arguments consistently return the same result as for masked arrays."""

        data1 = self.data1.astype(dtype1)
        data2 = self.data2.astype(dtype2)
        # If floating type, create NaNs
        if np.issubdtype(dtype1, np.floating):
            data1[self.mask1] = np.nan
        if np.issubdtype(dtype2, np.floating):
            data2[self.mask2] = np.nan

        pc1 = gu.PointCloud.from_xyz(x=self.coords[0], y=self.coords[1], z=data1, crs=self.crs)
        pc2 = gu.PointCloud.from_xyz(x=self.coords[0], y=self.coords[1], z=data2, crs=self.crs)

        ufunc = getattr(np, ufunc_str)

        # Find common dtype between the point cloud and the most constrained input type (first character is the input)
        try:
            com_dtype1 = np.promote_types(dtype1, ufunc.types[0][0])
        # The promote_types function raises an error for object dtypes (previously returned by find_common_dtypes)
        # (TypeError needed for backwards compatibility; also exceptions.DTypePromotionError for NumPy 1.25 and above)
        except TypeError:
            com_dtype1 = np.dtype("O")

        try:
            com_dtype2 = np.promote_types(dtype2, ufunc.types[0][1])
        # The promote_types function raises an error for object dtypes (previously returned by find_common_dtypes)
        # (TypeError needed for backwards compatibility; also exceptions.DTypePromotionError for NumPy 1.25 and above)
        except TypeError:
            com_dtype2 = np.dtype("O")

        # If the two input types can be the same type, pass a tuple with the common type of both
        # Below we ignore datetime and timedelta types "m" and "M", and int64 types "q" and "Q"
        if all(t[0] == t[1] for t in ufunc.types if not any(x in t[0:2] for x in ["m", "M", "q", "Q"])):
            try:
                com_dtype_both = np.promote_types(com_dtype1, com_dtype2)
            except TypeError:
                com_dtype_both = np.dtype("O")
            com_dtype_tuple = (com_dtype_both, com_dtype_both)

        # Otherwise, pass the tuple with each common type
        else:
            com_dtype_tuple = (com_dtype1, com_dtype2)

        # Catch warnings
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            # Check if both our input dtypes are possible on this ufunc, if yes check that outputs are identical
            if com_dtype_tuple in [(np.dtype(t[0]), np.dtype(t[1])) for t in ufunc.types]:  # noqa
                # For a single output
                if ufunc.nout == 1:
                    # There exists a single exception due to negative integers as exponent of integers in "power"
                    if ufunc_str == "power" and "int" in dtype1 and "int" in dtype2 and np.min(pc2.data) < 0:
                        with pytest.raises(ValueError, match="Integers to negative integer powers are not allowed."):
                            ufunc(pc1, pc2)
                        with pytest.raises(ValueError, match="Integers to negative integer powers are not allowed."):
                            ufunc(pc1.data, pc2.data)

                    # Otherwise, run the normal assertion for a single output
                    else:
                        assert np.array_equal(ufunc(pc1.data, pc2.data), ufunc(pc1, pc2).data, equal_nan=True)

                # For two outputs
                elif ufunc.nout == 2:
                    outputs_pc = ufunc(pc1, pc2)
                    output_arr = ufunc(pc1.data, pc2.data)
                    assert np.array_equal(output_arr[0], outputs_pc[0].data, equal_nan=True) and np.array_equal(
                        output_arr[1], outputs_pc[1].data, equal_nan=True
                    )

            # If the input dtype is not possible, check that NumPy raises a TypeError
            else:
                with pytest.raises(TypeError):
                    ufunc(pc1.data, pc2.data)
                with pytest.raises(TypeError):
                    ufunc(pc1, pc2)

    @pytest.mark.parametrize("arrfunc_str", handled_functions_1in)  # type: ignore
    @pytest.mark.parametrize("dtype", ["uint8", "int8", "float32"])  # type: ignore
    def test_array_functions_1nin(self, arrfunc_str: str, dtype: str) -> None:
        """
        Test that single-input array functions that we support give the same output as they would on the masked array.
        """

        data = self.data1.astype(dtype)
        # If floating type, create NaNs
        if np.issubdtype(dtype, np.floating):
            data[self.mask1] = np.nan
        pc = gu.PointCloud.from_xyz(x=self.coords[0], y=self.coords[1], z=data, crs=self.crs)

        # Get array func
        arrfunc = getattr(np, arrfunc_str)

        # Find the common dtype between the point cloud and the most constrained input type (first character is input)
        # com_dtype = np.find_common_type([dtype] + [arrfunc.types[0][0]], [])

        # Catch warnings
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)

            if "percentile" in arrfunc_str:
                args = (80.0,)
            elif "quantile" in arrfunc_str:
                args = (0.8,)
            else:
                args = ()  # type: ignore

            output_pc = arrfunc(pc, *args)
            output_arr = arrfunc(pc.data, *args)

            # This test is for when the NumPy function reduces the dimension of the array but not completely
            if isinstance(output_arr, np.ndarray):
                # When the NumPy function preserves the shape, it returns a point cloud
                if output_arr.shape == pc.data.shape:
                    assert isinstance(output_pc, gu.PointCloud)
                    assert np.array_equal(output_pc.data, output_arr, equal_nan=True)
                # Otherwise, it returns an array
                else:
                    assert np.array_equal(output_pc, output_arr, equal_nan=True)
            # This test is for when the NumPy function reduces the dimension to a single number
            else:
                assert np.array_equal(output_pc, output_arr, equal_nan=True)

    @pytest.mark.parametrize("arrfunc_str", handled_functions_2in)  # type: ignore
    @pytest.mark.parametrize("dtype1", ["uint8", "int8", "float32"])  # type: ignore
    @pytest.mark.parametrize("dtype2", ["uint8", "int8", "float32"])  # type: ignore
    def test_array_functions_2nin(self, arrfunc_str: str, dtype1: str, dtype2: str) -> None:
        """
        Test that double-input array functions that we support give the same output as they would on the masked array.
        """

        data1 = self.data1.astype(dtype1)
        data2 = self.data2.astype(dtype2)
        # If floating type, create NaNs
        if np.issubdtype(dtype1, np.floating):
            data1[self.mask1] = np.nan
        if np.issubdtype(dtype2, np.floating):
            data2[self.mask2] = np.nan

        pc1 = gu.PointCloud.from_xyz(x=self.coords[0], y=self.coords[1], z=data1, crs=self.crs)
        pc2 = gu.PointCloud.from_xyz(x=self.coords[0], y=self.coords[1], z=data2, crs=self.crs)

        # Get array func
        arrfunc = getattr(np, arrfunc_str)

        # Catch warnings
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)

            # Compute outputs
            output_pc = arrfunc(pc1, pc2)
            output_arr = arrfunc(pc1.data, pc2.data)

            # When the NumPy function preserves the shape, it returns a point cloud
            if isinstance(output_arr, np.ndarray) and output_arr.shape == pc1.data.shape:
                assert isinstance(output_pc, gu.PointCloud)
                assert np.array_equal(output_pc.data, output_arr, equal_nan=True)
            # Otherwise, it returns an array
            else:
                assert np.array_equal(output_pc, output_arr, equal_nan=True)

    @pytest.mark.parametrize("method_str", ["reduce"])  # type: ignore
    def test_ufunc_methods(self, method_str: str) -> None:
        """
        Test that universal function methods all behave properly, don't need to test all
        nodatas and dtypes as this was done above.
        """

        data2 = self.data2.astype("float32")
        data3 = self.data3.astype("float32")
        data1 = self.data1.astype("float32")
        data1[self.mask1] = np.nan
        data2[self.mask2] = np.nan
        data3[self.mask3] = np.nan

        pc1 = gu.PointCloud.from_xyz(x=self.coords[0], y=self.coords[1], z=data1, crs=self.crs)
        pc2 = gu.PointCloud.from_xyz(x=self.coords[0], y=self.coords[1], z=data2, crs=self.crs)
        pc3 = gu.PointCloud.from_xyz(x=self.coords[0], y=self.coords[1], z=data3, crs=self.crs)

        # Methods reduce, accumulate, reduceat and at only supported for binary function (2nin)
        # -- Test 1: -- Try a ufunc with 2nin, 1nout like np.add
        ufunc_2nin_1nout = getattr(np.add, method_str)
        output_pc = ufunc_2nin_1nout((pc1, pc2, pc3))
        output_arr = ufunc_2nin_1nout((pc1.data, pc2.data, pc3.data))

        assert np.array_equal(output_pc.data, output_arr, equal_nan=True)

        # Methods reduce only supports function that output a single value
        # -- Test 2: -- Try a ufunc with 2nin, 2nout: there's only divmod
        # ufunc_2nin_2nout = getattr(np.divmod, method_str)
        # outputs_pc = ufunc_2nin_2nout((rst1, rst2, rst3))
        # outputs_ma = ufunc_2nin_2nout((ma1, ma2, ma3))
        #
        # assert np.ma.allequal(outputs_ma[0], outputs_pc[0].data) and np.ma.allequal(
        #             outputs_ma[1], outputs_pc[1].data)

    @pytest.mark.parametrize(
        "np_func_name", ufuncs_str_2nin_1nout + ufuncs_str_2nin_2nout + handled_functions_2in
    )  # type: ignore
    def test_raise_errors_2nin(self, np_func_name: str) -> None:
        """Check that proper errors are raised when input pointcloud/array don't match (only 2-input functions)."""

        # Create point clouds
        pc = gu.PointCloud.from_xyz(x=self.coords[0], y=self.coords[1], z=self.data1, crs=self.crs)
        pc_wrong_shape = gu.PointCloud.from_array(self.arr_wrong_shape, crs=self.crs)
        pc_wrong_crs = gu.PointCloud.from_xyz(x=self.coords[0], y=self.coords[1], z=self.data1, crs=self.wrong_crs)
        pc_wrong_coords = gu.PointCloud.from_array(self.arr_wrong_coord, crs=self.crs)

        # Get ufunc
        np_func = getattr(np, np_func_name)

        # Strange errors happening only for these 4 functions...
        # See issue #457
        if np_func_name not in ["allclose", "isclose", "array_equal", "array_equiv"]:

            # Point clouds with different CRS or shape
            # Different shape
            georef_twopc_message = (
                "Both point clouds must have the same points X/Y coordinates and CRS for an arithmetic operation."
            )
            with pytest.raises(ValueError, match=re.escape(georef_twopc_message)):
                np_func(pc, pc_wrong_shape)

            with pytest.raises(ValueError, match=re.escape(georef_twopc_message)):
                np_func(pc, pc_wrong_coords)

            # Different CRS
            with pytest.raises(ValueError, match=re.escape(georef_twopc_message)):
                np_func(pc, pc_wrong_crs)

            # Array with different shape
            georef_pc_array_message = (
                "The array must be 1-dimensional with the same number of points as the point cloud for an arithmetic "
                "operation."
            )
            # Different shape, masked array
            # Check reflectivity just in case (just here, not later)
            with pytest.raises(ValueError, match=re.escape(georef_pc_array_message)):
                np_func(pc_wrong_shape.data, pc)
            with pytest.raises(ValueError, match=re.escape(georef_pc_array_message)):
                np_func(pc, pc_wrong_shape.data)



# Copyright (c) 2025 GeoUtils developers
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for dask-delayed functions for out-of-memory raster operations.
"""

from __future__ import annotations

import warnings
from typing import Any, Literal

import dask.array as da
import dask.delayed
import numpy as np
import rasterio as rio
from dask.utils import cached_cumsum
from scipy.interpolate import interpn

from geoutils._typing import NDArrayBool, NDArrayNum
from geoutils.raster.distributed_computing.chunked import (
    _build_geotiling_and_meta,
    _reproject_per_block,
)

# 1/ SUBSAMPLING
# At the date of April 2024:
# Getting an exact subsample size out-of-memory only for valid values is not supported directly by Dask/Xarray

# It is not trivial because we don't know where valid values will be in advance, and because of ragged output (varying
# output length considerations), which prevents from using high-level functions with good efficiency
# We thus follow https://blog.dask.org/2021/07/02/ragged-output (the dask.array.map_blocks solution has a larger RAM
# usage by having to drop an axis and re-chunk along 1D of the 2D array, so we use the dask.delayed solution instead)


def _get_subsample_size_from_user_input(
    subsample: int | float, total_nb_valids: int, silence_max_subsample: bool
) -> int:
    """Get subsample size based on a user input of either integer size or fraction of the number of valid points."""

    # If value is between 0 and 1, use a fraction
    if (subsample <= 1) & (subsample > 0):
        npoints = int(subsample * total_nb_valids)
    # Otherwise use the value directly
    elif subsample > 1:
        # Use the number of valid points if larger than subsample asked by user
        npoints = min(int(subsample), total_nb_valids)
        if subsample > total_nb_valids:
            if not silence_max_subsample:
                warnings.warn(
                    f"Subsample value of {subsample} is larger than the number of valid pixels of {total_nb_valids},"
                    f" using all valid pixels as a subsample.",
                    category=UserWarning,
                )
    else:
        raise ValueError("Subsample must be > 0.")

    return npoints


def _get_indices_block_per_subsample(
    indices_1d: NDArrayNum, num_chunks: tuple[int, int], nb_valids_per_block: list[int]
) -> list[list[int]]:
    """
    Get list of 1D valid subsample indices relative to the block for each block.

    The 1D valid subsample indices correspond to the subsample index to apply for a flattened array of valid values.
    Relative to the block means converted so that the block indexes for valid values starts at 0 up to the number of
    valid values in that block (while the input indices go from zero to the total number of valid values in the full
    array).

    :param indices_1d: Subsample 1D indexes among a total number of valid values.
    :param num_chunks: Number of chunks in X and Y.
    :param nb_valids_per_block: Number of valid pixels per block.

    :returns: Relative 1D valid subsample index per block.
    """

    # Apply a cumulative sum to get the first 1D total index of each block
    valids_cumsum = np.cumsum(nb_valids_per_block)

    # We can write a faster algorithm by sorting
    indices_1d = np.sort(indices_1d)

    # TODO: Write nested lists into array format to further save RAM?
    # We define a list of indices per block
    relative_index_per_block = [[] for _ in range(num_chunks[0] * num_chunks[1])]
    k = 0  # K is the block number
    for i in indices_1d:

        # Move to the next block K where current 1D subsample index is, if not in this one
        while i >= valids_cumsum[k]:
            k += 1

        # Add 1D subsample index  relative to first subsample index of this block
        first_index_block = valids_cumsum[k - 1] if k >= 1 else 0  # The first 1D valid subsample index of the block
        relative_index = i - first_index_block
        relative_index_per_block[k].append(relative_index)

    return relative_index_per_block


@dask.delayed  # type: ignore
def _delayed_nb_valids(arr_chunk: NDArrayNum | NDArrayBool) -> NDArrayNum:
    """Count number of valid values per block."""
    if arr_chunk.dtype == "bool":
        return np.array([np.count_nonzero(arr_chunk)]).reshape((1, 1))
    return np.array([np.count_nonzero(np.isfinite(arr_chunk))]).reshape((1, 1))


@dask.delayed  # type: ignore
def _delayed_subsample_block(
    arr_chunk: NDArrayNum | NDArrayBool, subsample_indices: NDArrayNum
) -> NDArrayNum | NDArrayBool:
    """Subsample the valid values at the corresponding 1D valid indices per block."""

    if arr_chunk.dtype == "bool":
        return arr_chunk[arr_chunk][subsample_indices]
    return arr_chunk[np.isfinite(arr_chunk)][subsample_indices]


@dask.delayed  # type: ignore
def _delayed_subsample_indices_block(
    arr_chunk: NDArrayNum | NDArrayBool, subsample_indices: NDArrayNum, block_id: dict[str, Any]
) -> NDArrayNum:
    """Return 2D indices from the subsampled 1D valid indices per block."""

    if arr_chunk.dtype == "bool":
        ix, iy = np.unravel_index(np.argwhere(arr_chunk.flatten())[subsample_indices], shape=arr_chunk.shape)
    else:
        #  Unravel indices of valid data to the shape of the block
        ix, iy = np.unravel_index(
            np.argwhere(np.isfinite(arr_chunk.flatten()))[subsample_indices], shape=arr_chunk.shape
        )

    # Convert to full-array indexes by adding the row and column starting indexes for this block
    ix += block_id["xstart"]
    iy += block_id["ystart"]

    return np.hstack((ix, iy))


def delayed_subsample(
    darr: da.Array,
    subsample: int | float = 1,
    return_indices: bool = False,
    random_state: int | np.random.Generator | None = None,
    silence_max_subsample: bool = False,
) -> NDArrayNum | tuple[NDArrayNum, NDArrayNum]:
    """
    Subsample a raster at valid values on out-of-memory chunks.

    Optionally, this function can return the 2D indices of the subsample of valid values instead.

    The random subsample is distributed evenly across valid values, no matter which chunk they belong to.
    First, the number of valid values in each chunk are computed out-of-memory. Then, a subsample is defined among
    the total number of valid values, which are then indexed sequentially along the chunk valid values out-of-memory.

    A random state will give a fixed subsample for a delayed array with a fixed chunksize. However, the subsample
    will vary with changing chunksize because the 1D delayed indexing depends on it (indexing per valid value per
    flattened chunk). For this reason, a loaded array will also have a different subsample due to its direct 1D
    indexing (per valid value for the entire flattened array).

    To ensure you reuse a similar subsample of valid values for several arrays, call this function with
    return_indices=True, then sample your arrays out-of-memory with .vindex[indices[0], indices[1]]
    (this assumes that these arrays have valid values at the same locations).

    Only valid values are sampled. If passing a numerical array, then only finite values are considered valid values.
    If passing a boolean array, then only True values are considered valid values.

    :param darr: Input dask array. This can be a boolean or a numerical array.
    :param subsample: Subsample size. If <= 1, will be considered a fraction of valid pixels to extract.
        If > 1 will be considered the number of valid pixels to extract.
    :param return_indices: If set to True, will return the extracted indices only.
    :param random_state: Random state, or seed number to use for random calculations.
    :param silence_max_subsample: Whether to silence the warning for the subsample size being larger than the total
        number of valid points (warns by default).

    :return: Subsample of values from the array (optionally, their indexes).
    """

    # Get random state
    rng = np.random.default_rng(random_state)

    # Compute number of valid points for each block out-of-memory
    blocks = darr.to_delayed().ravel()
    list_delayed_valids = [
        da.from_delayed(_delayed_nb_valids(b), shape=(1, 1), dtype=np.dtype("int32")) for b in blocks
    ]
    nb_valids_per_block = np.concatenate([dask.compute(*list_delayed_valids)]).squeeze()

    # Sum to get total number of valid points
    total_nb_valids = np.sum(nb_valids_per_block)

    # Get subsample size (depending on user input)
    subsample_size = _get_subsample_size_from_user_input(
        subsample=subsample, total_nb_valids=total_nb_valids, silence_max_subsample=silence_max_subsample
    )

    # Get random 1D indexes for the subsample size
    indices_1d = rng.choice(total_nb_valids, subsample_size, replace=False)

    # Sort which indexes belong to which chunk
    ind_per_block = _get_indices_block_per_subsample(
        indices_1d, num_chunks=darr.numblocks, nb_valids_per_block=nb_valids_per_block
    )

    # To just get the subsample without indices
    if not return_indices:
        # Task a delayed subsample to be computed for each block, skipping blocks with no values to sample
        list_subsamples = [
            _delayed_subsample_block(b, ind)
            for i, (b, ind) in enumerate(zip(blocks, ind_per_block))
            if len(ind_per_block[i]) > 0
        ]
        # Cast output to the right expected dtype and length, then compute and concatenate
        print(nb_valids_per_block)
        list_subsamples_delayed = [
            da.from_delayed(s, shape=(nb_valids_per_block[i],), dtype=darr.dtype) for i, s in enumerate(list_subsamples)
        ]
        subsamples = np.concatenate(dask.compute(*list_subsamples_delayed), axis=0)

        return subsamples

    # To return indices
    else:
        # Get starting 2D index for each chunk of the full array
        # (mirroring what is done in block_id of dask.array.map_blocks)
        # https://github.com/dask/dask/blob/24493f58660cb933855ba7629848881a6e2458c1/dask/array/core.py#L908
        # This list also includes the last index as well (not used here)
        starts = [cached_cumsum(c, initial_zero=True) for c in darr.chunks]
        num_chunks = darr.numblocks
        # Get the starts per 1D block ID by unravelling starting indexes for each block
        indexes_xi, indexes_yi = np.unravel_index(np.arange(len(blocks)), shape=(num_chunks[0], num_chunks[1]))
        block_ids = [
            {"xstart": starts[0][indexes_xi[i]], "ystart": starts[1][indexes_yi[i]]} for i in range(len(blocks))
        ]

        # Task delayed subsample indices to be computed for each block, skipping blocks with no values to sample
        list_subsample_indices = [
            _delayed_subsample_indices_block(b, ind, block_id=block_ids[i])
            for i, (b, ind) in enumerate(zip(blocks, ind_per_block))
            if len(ind_per_block[i]) > 0
        ]
        # Cast output to the right expected dtype and length, then compute and concatenate
        list_subsamples_indices_delayed = [
            da.from_delayed(s, shape=(2, len(ind_per_block[i])), dtype=np.dtype("int32"))
            for i, s in enumerate(list_subsample_indices)
        ]
        indices = np.concatenate(dask.compute(*list_subsamples_indices_delayed), axis=0)

        return indices[:, 0], indices[:, 1]


# 2/ POINT INTERPOLATION ON REGULAR OR EQUAL GRID
# At the date of April 2024:
# This functionality is not covered efficiently by Dask/Xarray, because they need to support rectilinear grids, which
# is difficult when interpolating in the chunked dimensions, and loads nearly all array memory when using .interp().

# Here we harness the fact that rasters are always on regular (or sometimes equal) grids to efficiently map
# the location of the blocks required for interpolation, which requires little memory usage.

# Code structure inspired by https://blog.dask.org/2021/07/02/ragged-output and the "block_id" in map_blocks


def _get_interp_indices_per_block(
    interp_x: NDArrayNum,
    interp_y: NDArrayNum,
    starts: list[tuple[int, ...]],
    num_chunks: tuple[int, int],
    chunksize: tuple[int, int],
    xres: float,
    yres: float,
) -> list[list[int]]:
    """Map blocks where each pair of interpolation coordinates will have to be computed."""

    # TODO 1: Check the robustness for chunksize different and X and Y

    # TODO 2: Check if computing block_i_id matricially + using an == comparison (possibly delayed) to get index
    #  per block is not more computationally efficient?
    #  (as it uses array instead of nested lists, and nested lists grow in RAM very fast)

    # The argument "starts" contains the list of chunk first X/Y index for the full array, plus the last index

    # We use one bucket per block, assuming a flattened blocks shape
    ind_per_block = [[] for _ in range(num_chunks[0] * num_chunks[1])]
    for i, (x, y) in enumerate(zip(interp_x, interp_y)):
        # Because it is a regular grid, we know exactly in which block ID the coordinate will fall
        block_i_1d = int((x - starts[0][0]) / (xres * chunksize[0])) * num_chunks[1] + int(
            (y - starts[1][0]) / (yres * chunksize[1])
        )
        ind_per_block[block_i_1d].append(i)

    return ind_per_block


@dask.delayed  # type: ignore
def _delayed_interp_points_block(
    arr_chunk: NDArrayNum, block_id: dict[str, Any], interp_coords: NDArrayNum
) -> NDArrayNum:
    """
    Interpolate block in 2D out-of-memory for a regular or equal grid.
    """

    # Extract information out of block_id dictionary
    xs, ys, xres, yres = (block_id["xstart"], block_id["ystart"], block_id["xres"], block_id["yres"])

    # Reconstruct the coordinates from xi/yi/xres/yres (as it has to be a regular grid)
    x_coords = np.arange(xs, xs + xres * arr_chunk.shape[0], xres)
    y_coords = np.arange(ys, ys + yres * arr_chunk.shape[1], yres)

    # TODO: Use scipy.map_coordinates for an equal grid as in Raster.interp_points?

    # Interpolate to points
    interp_chunk = interpn(points=(x_coords, y_coords), values=arr_chunk, xi=(interp_coords[0, :], interp_coords[1, :]))

    # And return the interpolated array
    return interp_chunk


def delayed_interp_points(
    darr: da.Array,
    points: tuple[list[float], list[float]],
    resolution: tuple[float, float],
    method: Literal["nearest", "linear", "cubic", "quintic"] = "linear",
) -> NDArrayNum:
    """
    Interpolate raster at point coordinates on out-of-memory chunks.

    This function harnesses the fact that a raster is defined on a regular (or equal) grid, and it is therefore
    faster than Xarray.interpn (especially for small sample sizes) and uses only a fraction of the memory usage.

    :param darr: Input dask array.
    :param points: Point(s) at which to interpolate raster value. If points fall outside of image, value
            returned is nan. Shape should be (N,2).
    :param resolution: Resolution of the raster (xres, yres).
    :param method: Interpolation method, one of 'nearest', 'linear', 'cubic', or 'quintic'. For more information,
            see scipy.ndimage.map_coordinates and scipy.interpolate.interpn. Default is linear.

    :return: Array of raster value(s) for the given points.
    """

    # TODO: Replace by a generic 2D point casting function accepting multiple inputs (living outside this function)
    # Convert input to 2D array
    points_arr = np.vstack((points[0], points[1]))

    # Map depth of overlap required for each interpolation method
    # TODO: Double-check this window somewhere in SciPy's documentation
    map_depth = {"nearest": 1, "linear": 2, "cubic": 3, "quintic": 5}

    # Expand dask array for overlapping computations
    chunksize = darr.chunksize
    expanded = da.overlap.overlap(darr, depth=map_depth[method], boundary=np.nan)

    # Get starting 2D index for each chunk of the full array
    # (mirroring what is done in block_id of dask.array.map_blocks)
    starts = [cached_cumsum(c, initial_zero=True) for c in darr.chunks]
    num_chunks = expanded.numblocks

    # Get samples indices per blocks
    ind_per_block = _get_interp_indices_per_block(
        points_arr[0, :], points_arr[1, :], starts, num_chunks, chunksize, resolution[0], resolution[1]
    )

    # Create a delayed object for each block, and flatten the blocks into a 1d shape
    blocks = expanded.to_delayed().ravel()

    # Build the block IDs by unravelling starting indexes for each block
    indexes_xi, indexes_yi = np.unravel_index(np.arange(len(blocks)), shape=(num_chunks[0], num_chunks[1]))
    block_ids = [
        {
            "xstart": (starts[0][indexes_xi[i]] - map_depth[method]) * resolution[0],
            "ystart": (starts[1][indexes_yi[i]] - map_depth[method]) * resolution[1],
            "xres": resolution[0],
            "yres": resolution[1],
        }
        for i in range(len(blocks))
    ]

    # Compute values delayed
    list_interp = [
        _delayed_interp_points_block(data_chunk, block_ids[i], points_arr[:, ind_per_block[i]])
        for i, data_chunk in enumerate(blocks)
        if len(ind_per_block[i]) > 0
    ]

    # We define the expected output shape and dtype to simplify things for Dask
    list_interp_delayed = [
        da.from_delayed(p, shape=(1, len(ind_per_block[i])), dtype=darr.dtype) for i, p in enumerate(list_interp)
    ]
    interp_points = np.concatenate(dask.compute(*list_interp_delayed), axis=0)

    # Re-order per-block output points to match their original indices
    indices = np.concatenate(ind_per_block).astype(int)
    argsort = np.argsort(indices)
    interp_points = np.array(interp_points)[argsort]

    return interp_points


# 3/ REPROJECT (see subfunctions in chunked module)


@dask.delayed  # type: ignore
def _delayed_reproject_per_block(
    *src_arrs: tuple[NDArrayNum], block_ids: list[dict[str, int]], combined_meta: dict[str, Any], **kwargs: Any
) -> NDArrayNum:
    """
    Delayed reprojection per destination block (also rebuilds a square array combined from intersecting source blocks).
    """
    return _reproject_per_block(*src_arrs, block_ids=block_ids, combined_meta=combined_meta, **kwargs)


def delayed_reproject(
    darr: da.Array,
    src_transform: rio.transform.Affine,
    src_crs: rio.crs.CRS,
    dst_transform: rio.transform.Affine,
    dst_shape: tuple[int, int],
    dst_crs: rio.crs.CRS,
    resampling: rio.enums.Resampling,
    src_nodata: int | float | None = None,
    dst_nodata: int | float | None = None,
    dst_chunksizes: tuple[int, int] | None = None,
    **kwargs: Any,
) -> da.Array:
    """
    Reproject georeferenced raster on out-of-memory chunks.

    Each chunk of the destination array is mapped to one or several intersecting chunks of the source array, and
    reprojection is performed using rio.warp.reproject for each mapping.

    Part of the code is inspired by https://github.com/opendatacube/odc-geo/pull/88.

    :param darr: Input dask array for source raster.
    :param src_transform: Geotransform of source raster.
    :param src_crs: Coordinate reference system of source raster.
    :param dst_transform: Geotransform of destination raster.
    :param dst_shape: Shape of destination raster.
    :param dst_crs: Coordinate reference system of destination raster.
    :param resampling: Resampling method.
    :param src_nodata: Nodata value of source raster.
    :param dst_nodata: Nodata value of destination raster.
    :param dst_chunksizes: Chunksizes for destination raster.
    :param kwargs: Other arguments to pass to rio.warp.reproject().

    :return: Dask array of reprojected raster.
    """

    # Define the chunking
    # For source, we can use the .chunks attribute
    src_chunks = darr.chunks

    if dst_chunksizes is None:
        dst_chunksizes = darr.chunksize
    # Prepare geotiling and reprojection metadata for source and destination grids
    src_geotiling, dst_geotiling, dst_chunks, dest2source, src_block_ids, meta_params, dst_block_geogrids = (
        _build_geotiling_and_meta(
            src_shape=darr.shape,
            src_transform=src_transform,
            src_crs=src_crs,
            dst_shape=dst_shape,
            dst_transform=dst_transform,
            dst_crs=dst_crs,
            src_chunks=src_chunks,
            dst_chunksizes=dst_chunksizes,
        )
    )

    # We call a delayed function that uses rio.warp to reproject the combined source block(s) to each destination block

    # Add fixed arguments to keywords
    kwargs.update(
        {
            "src_nodata": src_nodata,
            "dst_nodata": dst_nodata,
            "resampling": resampling,
            "src_crs": src_crs,
            "dst_crs": dst_crs,
        }
    )

    # Create a delayed object for each block, and flatten the blocks into a 1d shape
    blocks = darr.to_delayed().ravel()
    # Run the delayed reprojection, looping for each destination block
    list_reproj = [
        _delayed_reproject_per_block(
            *blocks[dest2source[i]], block_ids=meta_params[i][1], combined_meta=meta_params[i][0], **kwargs
        )
        for i in range(len(dest2source))
    ]

    # We define the expected output shape and dtype to simplify things for Dask
    list_reproj_delayed = [
        da.from_delayed(r, shape=dst_block_geogrids[i].shape, dtype=darr.dtype) for i, r in enumerate(list_reproj)
    ]

    # Array comes out as flat blocks x chunksize0 (varying) x chunksize1 (varying), so we can't reshape directly
    # We need to unravel the flattened blocks indices to align X/Y, then concatenate all columns, then rows
    indexes_xi, indexes_yi = np.unravel_index(
        np.arange(len(dest2source)), shape=(len(dst_chunks[0]), len(dst_chunks[1]))
    )

    lists_columns = [
        [l for i, l in enumerate(list_reproj_delayed) if j == indexes_xi[i]] for j in range(len(dst_chunks[0]))
    ]
    concat_columns = [da.concatenate(c, axis=1) for c in lists_columns]
    concat_all = da.concatenate(concat_columns, axis=0)

    return concat_all
